# -*- coding: utf-8 -*-
"""
Test pip_helpers.
"""

from __future__ import absolute_import, unicode_literals

import unittest

import django
import pytest

from ..pip_helpers import check_if_installed, get_installed_packages

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_pip_helpers"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestPipHelpers",)


@pytest.mark.django_db
class TestPipHelpers(unittest.TestCase):
    """Test pip_helpers."""

    @classmethod
    def setUpClass(cls):
        cls.mapping = {
            "country": {
                "name": "Netherlands",
                "province": {
                    "name": "North Holland",
                    "city": {
                        "name": "Amsterdam",
                    },
                },
            }
        }

    def test_get_installed_packages(self):
        """Test `get_installed_packages`.

        :return:
        """
        installed_packages = get_installed_packages()
        self.assertIn("Django", installed_packages)
        self.assertIn("elasticsearch", installed_packages)
        self.assertIn("elasticsearch-dsl", installed_packages)

    def test_get_installed_packages_with_versions(self):
        """Test `get_installed_packages`.

        :return:
        """
        installed_packages = get_installed_packages(with_versions=True)
        django_version = django.get_version()
        self.assertIn(("Django", django_version), installed_packages)

    def test_check_if_installed(self):
        """Test `check_if_installed`.

        :return:
        """
        self.assertTrue(check_if_installed("Django"))
        self.assertTrue(check_if_installed("elasticsearch"))
        self.assertTrue(check_if_installed("elasticsearch-dsl"))
        self.assertFalse(check_if_installed("django-fobi"))



# coding: utf-8
"""
Pagination.
"""

from __future__ import unicode_literals

from collections import OrderedDict

from django.core import paginator as django_paginator

from elasticsearch_dsl.utils import AttrDict

from rest_framework import pagination
from rest_framework.exceptions import NotFound
from rest_framework.response import Response

import six

from .versions import ELASTICSEARCH_GTE_6_0

__title__ = "django_elasticsearch_dsl_drf_alt.pagination"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "LimitOffsetPagination",
    "Page",
    "PageNumberPagination",
    "Paginator",
    "QueryFriendlyPageNumberPagination",
    "QueryFriendlyPaginator",
)


class GetCountMixin:

    def get_es_count(self, es_response):
        if isinstance(es_response, list):
            return len(es_response)
        if isinstance(es_response.hits.total, AttrDict):
            return es_response.hits.total.value
        return es_response.hits.total


class Page(django_paginator.Page, GetCountMixin):
    """Page for Elasticsearch."""

    def __init__(self, object_list, number, paginator, facets):
        self.facets = facets
        self.count = self.get_es_count(object_list)
        super(Page, self).__init__(object_list, number, paginator)


class Paginator(django_paginator.Paginator):
    """Paginator for Elasticsearch."""

    def page(self, number):
        """Returns a Page object for the given 1-based page number.

        :param number:
        :return:
        """
        number = self.validate_number(number)
        bottom = (number - 1) * self.per_page
        top = bottom + self.per_page
        if top + self.orphans >= self.count:
            top = self.count
        object_list = self.object_list[bottom:top].execute()
        __facets = getattr(object_list, "aggregations", None)
        return self._get_page(object_list, number, self, facets=__facets)

    def _get_page(self, *args, **kwargs):
        """Get page.

        Returns an instance of a single page.

        This hook can be used by subclasses to use an alternative to the
        standard :cls:`Page` object.
        """
        return Page(*args, **kwargs)


class QueryFriendlyPaginator(Paginator, GetCountMixin):
    """Paginator for Elasticsearch."""

    def page(self, number):
        """Returns a Page object for the given 1-based page number.

        :param number:
        :return:
        """
        bottom = (number - 1) * self.per_page
        top = bottom + self.per_page
        object_list = self.object_list[bottom:top].execute()
        self.count = int(self.get_es_count(object_list))
        if self.count > top and self.count - top <= self.orphans:
            # Fetch the additional orphaned nodes
            object_list = list(object_list) + list(
                self.object_list[top : self.count].execute()
            )
        number = self.validate_number(number)
        __facets = getattr(object_list, "aggregations", None)
        return self._get_page(object_list, number, self, facets=__facets)


class PageNumberPagination(pagination.PageNumberPagination, GetCountMixin):
    """Page number pagination.

    A simple page number based style that supports page numbers as
    query parameters.

    Example:

        http://api.example.org/accounts/?page=4
        http://api.example.org/accounts/?page=4&page_size=100
    """

    django_paginator_class = Paginator

    def __init__(self, *args, **kwargs):
        """Constructor.

        :param args:
        :param kwargs:
        """
        self.facets = None
        # self.page = None
        # self.request = None
        self.count = None
        super(PageNumberPagination, self).__init__(*args, **kwargs)

    def get_facets(self, page=None):
        """Get facets.

        :param page:
        :return:
        """
        if page is None:
            page = self.page

        if hasattr(page, "facets") and hasattr(page.facets, "_d_"):
            return page.facets._d_

    def paginate_queryset(self, queryset, request, view=None):
        """Paginate a queryset.

        Paginate a queryset if required, either returning a page object,
        or `None` if pagination is not configured for this view.

        :param queryset:
        :param request:
        :param view:
        :return:
        """
        # TODO: It seems that paginator breaks things. If take out, queries
        # doo work.
        # Check if there are suggest queries in the queryset,
        # ``execute_suggest`` method shall be called, instead of the
        # ``execute`` method and results shall be returned back immediately.
        # Placing this code at the very start of ``paginate_queryset`` method
        # saves us unnecessary queries.
        is_suggest = getattr(queryset, "_suggest", False)
        if is_suggest:
            if ELASTICSEARCH_GTE_6_0:
                return queryset.execute().to_dict().get("suggest")
            return queryset.execute_suggest().to_dict()

        # Check if we're using paginate queryset from `functional_suggest`
        # backend.
        if view.action == "functional_suggest":
            return queryset

        # If we got to this point, it means it's not a suggest or functional
        # suggest case.

        page_size = self.get_page_size(request)
        if not page_size:
            return None

        paginator = self.django_paginator_class(queryset, page_size)
        page_number = request.query_params.get(self.page_query_param, 1)
        if page_number in self.last_page_strings:
            page_number = paginator.num_pages

        # Something weird is happening here. If None returned before the
        # following code, post_filter works. If None returned after this code
        # post_filter does not work. Obviously, something strange happens in
        # the paginator.page(page_number) and thus affects the lazy
        # queryset in such a way, that we get TransportError(400,
        # 'parsing_exception', 'request does not support [post_filter]')
        try:
            self.page = paginator.page(page_number)
        except django_paginator.InvalidPage as exc:
            msg = self.invalid_page_message.format(
                page_number=page_number, message=six.text_type(exc)
            )
            raise NotFound(msg)

        if paginator.num_pages > 1 and self.template is not None:
            # The browsable API should display pagination controls.
            self.display_page_controls = True

        self.request = request
        return list(self.page)

    def get_paginated_response_context(self, data):
        """Get paginated response data.

        :param data:
        :return:
        """
        __data = [
            ("count", self.page.count),
            # ('count', self.count),
            ("next", self.get_next_link()),
            ("previous", self.get_previous_link()),
        ]
        __facets = self.get_facets()
        if __facets is not None:
            __data.append(
                ("facets", __facets),
            )
        __data.append(
            ("results", data),
        )
        return __data

    def get_paginated_response(self, data):
        """Get paginated response.

        :param data:
        :return:
        """
        return Response(OrderedDict(self.get_paginated_response_context(data)))


class QueryFriendlyPageNumberPagination(PageNumberPagination):
    """Page number pagination.

    A simple page number based style that supports page numbers as
    query parameters.

    Example:

        http://api.example.org/accounts/?page=4
        http://api.example.org/accounts/?page=4&page_size=100
    """

    django_paginator_class = QueryFriendlyPaginator
    page_size_query_param = "page_size"
    orphans_query_param = "orphans"

    def paginate_queryset(self, queryset, request, view=None):
        """Paginate a queryset.

        Paginate a queryset if required, either returning a page object,
        or `None` if pagination is not configured for this view.

        :param queryset:
        :param request:
        :param view:
        :return:
        """
        # TODO: It seems that paginator breaks things. If take out, queries
        # doo work.
        # Check if there are suggest queries in the queryset,
        # ``execute_suggest`` method shall be called, instead of the
        # ``execute`` method and results shall be returned back immediately.
        # Placing this code at the very start of ``paginate_queryset`` method
        # saves us unnecessary queries.
        is_suggest = getattr(queryset, "_suggest", False)
        if is_suggest:
            if ELASTICSEARCH_GTE_6_0:
                return queryset.execute().to_dict().get("suggest")
            return queryset.execute_suggest().to_dict()

        # Check if we're using paginate queryset from `functional_suggest`
        # backend.
        if view.action == "functional_suggest":
            return queryset

        # If we got to this point, it means it's not a suggest or functional
        # suggest case.

        page_size = self.get_page_size(request)
        if not page_size:
            return None

        orphans = min(
            int(request.query_params.get(self.orphans_query_param, 0)), page_size
        )
        paginator = self.django_paginator_class(queryset, page_size, orphans=orphans)
        page_number = int(request.query_params.get(self.page_query_param, 1))
        if page_number in self.last_page_strings:
            page_number = paginator.num_pages

        # Something weird is happening here. If None returned before the
        # following code, post_filter works. If None returned after this code
        # post_filter does not work. Obviously, something strange happens in
        # the paginator.page(page_number) and thus affects the lazy
        # queryset in such a way, that we get TransportError(400,
        # 'parsing_exception', 'request does not support [post_filter]')
        try:
            self.page = paginator.page(page_number)
        except django_paginator.InvalidPage as exc:
            msg = self.invalid_page_message.format(
                page_number=page_number, message=six.text_type(exc)
            )
            raise NotFound(msg)

        if paginator.num_pages > 1 and self.template is not None:
            # The browsable API should display pagination controls.
            self.display_page_controls = True

        self.request = request
        return list(self.page)


class LimitOffsetPagination(pagination.LimitOffsetPagination, GetCountMixin):
    """A limit/offset pagination.

    Example:

        http://api.example.org/accounts/?limit=100
        http://api.example.org/accounts/?offset=400&limit=100
    """

    def __init__(self, *args, **kwargs):
        """Constructor.

        :param args:
        :param kwargs:
        """
        self.facets = None
        self.count = None
        # self.limit = None
        # self.offset = None
        # self.request = None
        super(LimitOffsetPagination, self).__init__(*args, **kwargs)

    def paginate_queryset(self, queryset, request, view=None):
        # Check if there are suggest queries in the queryset,
        # ``execute_suggest`` method shall be called, instead of the
        # ``execute`` method and results shall be returned back immediately.
        # Placing this code at the very start of ``paginate_queryset`` method
        # saves us unnecessary queries.
        is_suggest = getattr(queryset, "_suggest", False)
        if is_suggest:
            if ELASTICSEARCH_GTE_6_0:
                return queryset.execute().to_dict().get("suggest")
            return queryset.execute_suggest().to_dict()

        # Check if we're using paginate queryset from `functional_suggest`
        # backend.
        if view.action == "functional_suggest":
            return queryset

        # If we got to this point, it means it's not a suggest or functional
        # suggest case.

        # if hasattr(self, 'get_es_count'):
        #     self.count = self.get_es_count(queryset)
        # else:
        #     from rest_framework.pagination import _get_count
        #     self.count = _get_count(queryset)

        self.limit = self.get_limit(request)
        if self.limit is None:
            return None

        self.offset = self.get_offset(request)
        self.request = request

        resp = queryset[self.offset : self.offset + self.limit].execute()
        self.facets = getattr(resp, "aggregations", None)

        self.count = self.get_es_count(resp)

        if self.count > self.limit and self.template is not None:
            self.display_page_controls = True

        if self.count == 0 or self.offset > self.count:
            return []
        return list(resp)

    def get_facets(self, facets=None):
        """Get facets.

        :param facets:
        :return:
        """
        if facets is None:
            facets = self.facets

        if facets is None:
            return None

        if hasattr(facets, "_d_"):
            return facets._d_

    def get_paginated_response_context(self, data):
        """Get paginated response data.

        :param data:
        :return:
        """
        __data = [
            ("count", self.count),
            ("next", self.get_next_link()),
            ("previous", self.get_previous_link()),
        ]
        __facets = self.get_facets()
        if __facets is not None:
            __data.append(
                ("facets", __facets),
            )
        __data.append(
            ("results", data),
        )
        return __data

    def get_paginated_response(self, data):
        """Get paginated response.

        :param data:
        :return:
        """
        return Response(OrderedDict(self.get_paginated_response_context(data)))



"""Base search backend."""

from django.core.exceptions import ImproperlyConfigured

from django_elasticsearch_dsl import fields
from rest_framework.filters import BaseFilterBackend
from rest_framework.settings import api_settings

from ..mixins import FilterBackendMixin
from ...compat import coreapi, coreschema
from ...constants import MATCHING_OPTIONS, DEFAULT_MATCHING_OPTION

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.search.common"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("BaseSearchFilterBackend",)


class BaseSearchFilterBackend(BaseFilterBackend, FilterBackendMixin):
    """Base search filter backend."""

    query_backends = []

    matching = DEFAULT_MATCHING_OPTION

    search_param = api_settings.SEARCH_PARAM

    def get_search_query_params(self, request):
        """Get search query params.

        :param request: Django REST framework request.
        :type request: rest_framework.request.Request
        :return: List of search query params.
        :rtype: list
        """
        query_params = request.query_params.copy()
        return query_params.getlist(self.search_param, [])

    def get_query_backends(self, request, view):
        """Get query backends.

        :return:
        """
        raise NotImplementedError(
            "You should define `get_query_backends` method in your {} class"
            "".format(self.__class__.__name__)
        )

    def _get_query_backends(self, request, view):
        """Get query backends internal.

        :param request:
        :param view:
        :return:
        """
        try:
            return self.get_query_backends(request, view)
        except NotImplementedError as err:
            pass

        if not self.query_backends:
            raise NotImplementedError(
                "Your search backend shall either implement "
                "`get_query_backends` method or define `query_backends`"
                "property."
            )
        return self.query_backends[:]

    def filter_queryset(self, request, queryset, view):
        """Filter the queryset.

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        if self.matching not in MATCHING_OPTIONS:
            raise ImproperlyConfigured(
                "Your `matching` value does not match the allowed matching"
                "options: {}".format(", ".join(MATCHING_OPTIONS))
            )

        __query_backends = self._get_query_backends(request, view)

        if len(__query_backends) > 1:
            __queries = []
            for query_backend in __query_backends:
                __queries.extend(
                    query_backend.construct_search(
                        request=request, view=view, search_backend=self
                    )
                )

            if __queries:
                queryset = queryset.query("bool", **{self.matching: __queries})

        elif len(__query_backends) == 1:
            __query = __query_backends[0].construct_search(
                request=request, view=view, search_backend=self
            )
            queryset = queryset.query("bool", **{self.matching: __query})

        else:
            raise ImproperlyConfigured(
                "Search filter backend shall have at least one query_backend"
                "specified either in `query_backends` property or "
                "`get_query_backends` method. Make appropriate changes to"
                "your {} class".format(self.__class__.__name__)
            )

        return queryset

    def get_coreschema_field(self, field):
        if isinstance(field, fields.IntegerField):
            field_cls = coreschema.Number
        else:
            field_cls = coreschema.String
        return field_cls()

    def get_schema_fields(self, view):
        assert coreapi is not None, (
            "coreapi must be installed to " "use `get_schema_fields()`"
        )
        assert coreschema is not None, (
            "coreschema must be installed to " "use `get_schema_fields()`"
        )

        _search_fields = getattr(view, "search_fields", None)
        if isinstance(_search_fields, dict):
            search_fields = list(_search_fields.keys())
        else:
            search_fields = _search_fields

        return (
            []
            if not search_fields
            else [
                coreapi.Field(
                    name=self.search_param,
                    required=False,
                    location="query",
                    schema=coreschema.String(
                        description="Search in " "{}.".format(", ".join(search_fields))
                    ),
                )
            ]
        )



"""
Ids filtering backend.

Filters documents that only have the provided ids. Note, this query uses the
`_uid` field.

Elastic query:

    {
        "query": {
            "ids": {
                "type": "book_document",
                "values": ["68", "64", "58"]
            }
        }
    }

REST framework request equivalent:

- http://localhost:8000/api/articles/?ids=68__64__58
- http://localhost:8000/api/articles/?ids=68&ids=64&ids=58

Official Elastic docs:

- https://www.elastic.co/guide/en/elasticsearch/reference/current/
  query-dsl-ids-query.html
"""

from rest_framework.filters import BaseFilterBackend

from ...versions import ELASTICSEARCH_LTE_6_0
from ..mixins import FilterBackendMixin

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.filtering.ids"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("IdsFilterBackend",)


class IdsFilterBackend(BaseFilterBackend, FilterBackendMixin):
    """Ids filter backend for Elasticsearch.

    Example:

        >>> from django_elasticsearch_dsl_drf_alt.filter_backends import (
        >>>     IdsFilterBackend
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.viewsets import (
        >>>     BaseDocumentViewSet,
        >>> )
        >>>
        >>> # Local article document definition
        >>> from .documents import ArticleDocument
        >>>
        >>> # Local article document serializer
        >>> from .serializers import ArticleDocumentSerializer
        >>>
        >>> class ArticleDocumentView(BaseDocumentViewSet):
        >>>
        >>>     document = ArticleDocument
        >>>     serializer_class = ArticleDocumentSerializer
        >>>     filter_backends = [IdsFilterBackend]
    """

    ids_query_param = "ids"

    def get_ids_query_params(self, request):
        """Get search query params.

        :param request: Django REST framework request.
        :type request: rest_framework.request.Request
        :return: List of search query params.
        :rtype: list
        """
        query_params = request.query_params.copy()
        return query_params.getlist(self.ids_query_param, [])

    def get_ids_values(self, request, view):
        """Get ids values for query.

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        query_params = self.get_ids_query_params(request)
        _ids = []
        for _id in query_params:
            _values = self.split_lookup_complex_value(_id)
            _ids += _values
        return _ids

    def filter_queryset(self, request, queryset, view):
        """Filter the queryset.

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        __ids = self.get_ids_values(request, view)
        if __ids:
            _ids = [_i for _i in __ids if _i]
            _qs_kwargs = {"values": _ids}
            # Prior 7.x ``type`` argument was accepted. Starting from 7.x
            # it has been deprecated. As long as 6.x is supported, this
            # should stay.
            if ELASTICSEARCH_LTE_6_0:
                _qs_kwargs.update({"type": view.mapping})

            queryset = queryset.query("ids", **_qs_kwargs)

        return queryset



# -*- coding: utf-8 -*-
"""
Test helpers.
"""

from __future__ import absolute_import, unicode_literals

import unittest

from django.core.management import call_command

import pytest

import factories

from ..helpers import more_like_this

from .base import BaseTestCase

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_helpers"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestHelpers",)


@pytest.mark.django_db
class TestHelpers(BaseTestCase):
    """Test helpers."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        super(TestHelpers, cls).setUpClass()
        cls.books_count = 30
        cls.books = []
        cls.books.append(
            factories.BookWithoutTagsFactory(
                title="Magic Of Thinking Big",
                description="Millions of people throughout the world have "
                "improved their lives using The Magic of "
                "Thinking Big. Dr. David J. Schwartz, long "
                "regarded as one of the foremost experts on "
                "motivation, will help you sell better, manage "
                "better, earn more money, and-most important "
                "of all-find greater happiness and peace of mind.",
            )
        )
        cls.books.append(
            factories.BookWithoutTagsFactory(
                title="The Power of Positive Thinking",
                description="The book describes the power positive thinking "
                "has and how a firm belief in something, does "
                "actually help in achieving it.",
            )
        )
        cls.books.append(
            factories.BookWithoutTagsFactory(
                title="Think and Grow Rich",
                description="Think And Grow Rich has earned itself the "
                "reputation of being considered a textbook for "
                "actionable techniques that can help one get "
                "better at doing anything, not just by rich and "
                "wealthy, but also by people doing wonderful work "
                "in their respective fields.",
            )
        )
        cls.books.append(
            factories.BookWithoutTagsFactory(
                title="The Magic of thinking Big",
                description="First published in 1959, David J Schwartz's "
                "classic teachings are as powerful today as they "
                "were then. Practical, empowering and hugely "
                "engaging, this book will not only inspire you, "
                "it will give you the tools to change your life "
                "for the better - starting from now.",
            )
        )
        cls.books.append(
            factories.BookWithoutTagsFactory(
                title="How to Stop Worrying and Start Living",
                description="The book is written to help readers by changing "
                "their habit of worrying. The author Dale "
                "Carnegie has shared his personal experiences, "
                "wherein he was mostly unsatisfied and worried "
                "about lot of life situations.",
            )
        )
        cls.books.append(
            factories.BookWithoutTagsFactory(
                title="Practicing The Power Of Now",
                description="To make the journey into The Power of Now we "
                "will need to leave our analytical mind and its "
                "false created self, the ego, behind.",
            )
        )

        cls.sleep()
        call_command("search_index", "--rebuild", "-f")

    def _more_like_this(self, obj, fields):
        """Filter by field."""
        res = more_like_this(obj, fields, 2, 2, 2, 10)
        self.assertGreater(res.count(), 0)

    def test_filter_by_field(self):
        """Filter by field."""
        obj = self.books[0]
        return self._more_like_this(obj, ["title", "description"])


if __name__ == "__main__":
    unittest.main()



"""
Test multi match search filter backend.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

from books import constants
import factories
from search_indexes.viewsets import BookMultiMatchSearchFilterBackendDocumentViewSet
from ..filter_backends import MultiMatchSearchFilterBackend

from .base import (
    BaseRestFrameworkTestCase,
    CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED,
    CORE_API_AND_CORE_SCHEMA_MISSING_MSG,
)

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_search_multi_match"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestMultiMatchSearch",)


@pytest.mark.django_db
class TestMultiMatchSearch(BaseRestFrameworkTestCase):
    """Test multi match search."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        super(TestMultiMatchSearch, cls).setUpClass()

        # Book factories with unique title
        cls.special_count = 10
        cls.special = factories.BookWithUniqueTitleFactory.create_batch(
            cls.special_count,
            **{
                "summary": "Delusional Insanity, fine art photography",
                "state": constants.BOOK_PUBLISHING_STATUS_PUBLISHED,
            }
        )

        # Lorem ipsum book factories
        cls.lorem_count = 10
        cls.lorem = factories.BookWithUniqueTitleFactory.create_batch(cls.lorem_count)

        # Book factories with title, description and summary that actually
        # make sense
        cls.non_lorem_count = 9
        cls.non_lorem = [
            factories.BookChapter20Factory(),
            factories.BookChapter21Factory(),
            factories.BookChapter22Factory(),
            factories.BookChapter60Factory(),
            factories.BookChapter61Factory(),
            factories.BookChapter62Factory(),
            factories.BookChapter110Factory(),
            factories.BookChapter111Factory(),
            factories.BookChapter112Factory(),
        ]

        cls.all_count = cls.special_count + cls.lorem_count + cls.non_lorem_count

        cls.cities_count = 20
        cls.cities = factories.CityFactory.create_batch(cls.cities_count)

        # Create 10 cities in a given country. The reason that we don't
        # do create_batch here is that sometimes in the same test city name is
        # generated twice and thus our concept of precise number matching
        # fails. Before there's a better approach, this would stay so. The
        # create_batch part (below) will remain commented out, until there's a
        # better solution.
        cls.switzerland = factories.CountryFactory.create(name="Wonderland")
        cls.switz_cities_count = 10
        cls.switz_cities_names = [
            "Zurich",
            "Geneva",
            "Basel",
            "Lausanne",
            "Bern",
            "Winterthur",
            "Lucerne",
            "St. Gallen",
            "Lugano",
            "Biel/Bienne",
        ]
        for switz_city in cls.switz_cities_names:
            cls.switz_cities = factories.CityFactory(
                name=switz_city, country=cls.switzerland
            )
        # cls.switz_cities = factories.CityFactory.create_batch(
        #     cls.switz_cities_count,
        #     country=cls.switzerland
        # )
        cls.all_cities_count = cls.cities_count + cls.switz_cities_count

        cls.sleep()
        call_command("search_index", "--rebuild", "-f")

        # Testing coreapi and coreschema
        cls.backend = MultiMatchSearchFilterBackend()
        cls.view = BookMultiMatchSearchFilterBackendDocumentViewSet()

    def _search(self, search_term, num_results, url=None):
        """Search."""
        self.authenticate()

        if url is None:
            url = reverse("bookdocument_multi_match_search_backend-list", kwargs={})

        data = {}

        # Should contain 20 results
        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), self.all_count)

        # Should contain only 10 results
        filtered_response = self.client.get(
            url + "?search_multi_match={}".format(search_term), data
        )
        self.assertEqual(filtered_response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(filtered_response.data["results"]), num_results)

    def _search_boost(self, search_term, ordering, url=None):
        """Search boost.

        In our book view, we have the following defined:

        >>> search_fields = {
        >>>     'title': {'boost': 4},
        >>>     'description': {'boost': 2},
        >>>     'summary': None,
        >>> }

        That means that `title` is more important than `description` and
        `description` is more important than `summary`.
        Results with search term in `title`, `summary` and
        `description` shall be ranked better than results with search term
        in `summary` and `description`. In their turn, results with search term
        in `summary` and `description` shall be ranked better than results
        with search term in `description` only.

        :param search_term:
        :param ordering:
        :return:
        """
        self.authenticate()

        if url is None:
            url = reverse(
                "bookdocument_multi_match_phrase_prefix_search_backend-list", kwargs={}
            )
        data = {}

        filtered_response = self.client.get(
            url + "?search_multi_match={}".format(search_term), data
        )
        self.assertEqual(filtered_response.status_code, status.HTTP_200_OK)
        self.assertIn("results", filtered_response.data)
        for counter, item_id in enumerate(ordering):
            result_item = filtered_response.data["results"][counter]
            self.assertEqual(result_item["id"], item_id)

    def test_search(self, url=None):
        """Search."""
        return self._search(search_term="Pool of Tears", num_results=3, url=url)

    def test_search_boost(self, url=None):
        """Search boost.

        :return:
        """
        # Search for: The Pool of Tears
        self._search_boost(
            search_term="The Pool of Tears",
            ordering=[
                self.non_lorem[0].pk,
                self.non_lorem[1].pk,
                self.non_lorem[2].pk,
            ],
            url=url,
        )

        # Search for: Pig and Pepper
        self._search_boost(
            search_term="Pig and Pepper",
            ordering=[
                self.non_lorem[3].pk,
            ],
            url=url,
        )

        # Search for: and Pepper"
        self._search_boost(
            search_term="and Pepper",
            ordering=[
                self.non_lorem[3].pk,
                self.non_lorem[4].pk,
                self.non_lorem[5].pk,
            ],
            url=url,
        )

        # Search for: Who Stole the Tarts
        self._search_boost(
            search_term="Who Stole the Tarts",
            ordering=[
                self.non_lorem[6].pk,
            ],
            url=url,
        )

        # Search for: Stole the Tarts
        self._search_boost(
            search_term="Stole the Tarts",
            ordering=[
                self.non_lorem[6].pk,
                self.non_lorem[7].pk,
            ],
            url=url,
        )

        # Search for: the Tarts
        self._search_boost(
            search_term="the Tarts",
            ordering=[
                self.non_lorem[6].pk,
                self.non_lorem[7].pk,
                self.non_lorem[8].pk,
            ],
            url=url,
        )

    def test_search_selected_fields(self, url=None):
        """Search boost.

        :return:
        """
        # Search for (only in `title` and `summary`): and Pepper
        self._search(search_term="title,summary:and Pepper", num_results=2, url=url)

    def test_search_boost_selected_fields(self, url=None):
        """Search boost.

        :return:
        """
        # Search for (only in `title` and `summary`): and Pepper
        self._search_boost(
            search_term="title,summary:and Pepper",
            ordering=[
                self.non_lorem[3].pk,
                self.non_lorem[4].pk,
            ],
            url=url,
        )

    @unittest.skipIf(
        not CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED, CORE_API_AND_CORE_SCHEMA_MISSING_MSG
    )
    def test_schema_fields_with_filter_fields_list(self):
        """Test schema field generator"""
        fields = self.backend.get_schema_fields(self.view)
        fields = [f.name for f in fields]
        self.assertEqual(fields, ["search"])

    @unittest.skipIf(
        not CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED, CORE_API_AND_CORE_SCHEMA_MISSING_MSG
    )
    def test_schema_field_not_required(self):
        """Test schema fields always not required"""
        fields = self.backend.get_schema_fields(self.view)
        fields = [f.required for f in fields]
        for field in fields:
            self.assertFalse(field)


if __name__ == "__main__":
    unittest.main()



"""
Analyzers.
"""

from elasticsearch_dsl import analyzer
from elasticsearch_dsl.analysis import token_filter

__title__ = "django_elasticsearch_dsl_drf_alt.analyzers"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "edge_ngram_completion_filter",
    "edge_ngram_completion",
)

edge_ngram_completion_filter = token_filter(
    "edge_ngram_completion_filter", type="edge_ngram", min_gram=1, max_gram=20
)


edge_ngram_completion = analyzer(
    "edge_ngram_completion",
    tokenizer="standard",
    filter=["lowercase", edge_ngram_completion_filter],
)



"""
Ordering backends.
"""

from .common import DefaultOrderingFilterBackend, OrderingFilterBackend
from .geo_spatial import GeoSpatialOrderingFilterBackend

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.ordering"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "DefaultOrderingFilterBackend",
    "GeoSpatialOrderingFilterBackend",
    "OrderingFilterBackend",
)



from django.conf import settings
from elasticsearch import Elasticsearch

__all__ = (
    "get_all_indices",
    "delete_all_indices",
)


def get_all_indices(with_protected=False):
    """Get all indices.
    Args:
        with_protected (bool):
    Returns:
        list: List of indices.
    """
    es = Elasticsearch(**settings.ELASTICSEARCH_DSL["default"])
    _indices = es.indices.get_alias("*").items()
    if with_protected:
        return [_i for _i, _o in _indices]
    else:
        return [_i for _i, _o in _indices if not _i.startswith(".")]


def delete_all_indices(with_protected=False):
    """Delete all indices.
    Args:
        with_protected (bool):
    Returns:
        tuple: Tuple of two lists with removed and errored indices.
    """
    es = Elasticsearch(**settings.ELASTICSEARCH_DSL["default"])
    _indices = get_all_indices(with_protected=with_protected)
    _ok = []
    _fail = []
    for _i in _indices:
        try:
            _res = es.indices.delete(_i)
        except Exception as err:
            _fail.append(_i)
        if _res and isinstance(_res, dict) and _res.get("acknowledged", False):
            _ok.append(_i)
    return _ok, _fail



"""
Test geo-spatial ordering filter backend.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

import factories

from ..constants import (
    GEO_DISTANCE_ORDERING_PARAM,
    SEPARATOR_LOOKUP_COMPLEX_VALUE,
)
from .base import BaseRestFrameworkTestCase

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_filtering"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestOrderingGeoSpatial",)


@pytest.mark.django_db
class TestOrderingGeoSpatial(BaseRestFrameworkTestCase):
    """Test ordering geo-spatial."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up."""
        super(TestOrderingGeoSpatial, cls).setUpClass()

        cls.geo_origin = factories.PublisherFactory.create(
            **{
                "latitude": 48.8549,
                "longitude": 2.3000,
            }
        )

        cls.geo_in_count = 5
        cls.unit = "km"
        cls.algo = "plane"
        cls.geo_in = []
        for index in range(cls.geo_in_count):
            __publisher = factories.PublisherFactory.create(
                **{
                    "latitude": 48.8570 + index,
                    "longitude": 2.3005,
                }
            )
            cls.geo_in.append(__publisher)

        cls.base_publisher_url = reverse("publisherdocument-list", kwargs={})

        cls.sleep()
        call_command("search_index", "--rebuild", "-f")

    @pytest.mark.webtest
    def test_field_filter_geo_distance(self):
        """Field filter geo_distance.

        Example:

            http://localhost:8000
            /api/publisher/?ordering=location;48.85;2.30;km;plane
        """
        self.authenticate()

        __params = "location{sep}{lat}{sep}{lon}{sep}{unit}{sep}{algo}".format(
            lat=self.geo_origin.latitude,
            lon=self.geo_origin.longitude,
            unit=self.unit,
            algo=self.algo,
            sep=SEPARATOR_LOOKUP_COMPLEX_VALUE,
        )

        url = self.base_publisher_url[:] + "?{}={}".format(
            GEO_DISTANCE_ORDERING_PARAM, __params
        )

        data = {}
        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # Should contain only 6 results
        self.assertEqual(len(response.data["results"]), self.geo_in_count + 1)
        item_count = len(response.data["results"])

        for counter, item in enumerate(response.data["results"]):
            if (counter > 1) and (counter < item_count + 1):
                self.assertLess(
                    response.data["results"][counter - 1]["location"]["lat"],
                    response.data["results"][counter]["location"]["lat"],
                )


if __name__ == "__main__":
    unittest.main()



"""
Test filtering `post_filter` backend.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

from search_indexes.viewsets import BookDocumentViewSet

from ..filter_backends import PostFilterFilteringFilterBackend
from .base import BaseRestFrameworkTestCase
from .data_mixins import AddressesMixin, BooksMixin

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_filtering_post"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestFilteringGlobalAggregations",)


@pytest.mark.django_db
class TestFilteringGlobalAggregations(
    BaseRestFrameworkTestCase, AddressesMixin, BooksMixin
):
    """Test filtering with global aggregations."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up."""
        super(TestFilteringGlobalAggregations, cls).setUpClass()
        # Testing simple documents: Publisher index.
        cls.create_books()

        # Testing nested objects: Addresses, cities and countries
        cls.created_addresses()

        cls.sleep()
        # Update the Elasticsearch index
        call_command("search_index", "--rebuild", "-f")

        # Testing coreapi and coreschema
        cls.backend = PostFilterFilteringFilterBackend()
        cls.view = BookDocumentViewSet()

    # ***********************************************************************
    # ************************ Test global facets ***************************
    # ***********************************************************************

    # This is what's it's all about - the facets.
    def _list_results_with_facets(self):
        """List results with facets."""
        self.authenticate()

        url = reverse("bookdocument-list", kwargs={})
        facet_state_url = url + "?facet=state_global"
        data = {}

        # ******************************************************************
        # ************************ No args response ************************
        # ******************************************************************

        # Make request
        no_args_response = self.client.get(url, data)
        self.assertEqual(no_args_response.status_code, status.HTTP_200_OK)

        # Should contain `self.all_count` results
        self.assertEqual(len(no_args_response.data["results"]), self.all_count)

        # Should contain 1 facets
        self.assertEqual(len(no_args_response.data["facets"]), 1)

        # ******************************************************************
        # ********************** With facets response **********************
        # ******************************************************************

        # Make request
        facet_state_response = self.client.get(facet_state_url, data)
        self.assertEqual(facet_state_response.status_code, status.HTTP_200_OK)

        # Should contain `self.all_count` results
        self.assertEqual(len(facet_state_response.data["results"]), self.all_count)

        # Should contain 2 facets
        self.assertEqual(len(facet_state_response.data["facets"]), 2)
        # With 3 statuses
        self.assertEqual(
            len(
                facet_state_response.data["facets"]["_filter_state_global"][
                    "state_global"
                ]["buckets"]
            ),
            3,
        )

        self.assertIn("_filter_state_global", facet_state_response.data["facets"])
        self.assertIn(
            "state_global", facet_state_response.data["facets"]["_filter_state_global"]
        )
        self.assertIn(
            "buckets",
            facet_state_response.data["facets"]["_filter_state_global"]["state_global"],
        )
        # self.assertIn(
        #     'buckets',
        #     facet_state_response.data['facets']['_filter_state']['state']
        # )
        self.assertIn(
            {"doc_count": 10, "key": "published"},
            facet_state_response.data["facets"]["_filter_state_global"]["state_global"][
                "buckets"
            ],
        )
        self.assertIn(
            {"doc_count": 10, "key": "in_progress"},
            facet_state_response.data["facets"]["_filter_state_global"]["state_global"][
                "buckets"
            ],
        )
        self.assertIn(
            {"doc_count": 7, "key": "rejected"},
            facet_state_response.data["facets"]["_filter_state_global"]["state_global"][
                "buckets"
            ],
        )

        # ******************************************************************
        # ******************* With facets filtered response ****************
        # ******************************************************************

        facet_state_filtered_url = url + "?facet=state_global&state=published"
        # Make request
        facet_state_filtered_response = self.client.get(facet_state_filtered_url, data)
        self.assertEqual(facet_state_filtered_response.status_code, status.HTTP_200_OK)

        # Should contain `self.published_count` results
        self.assertEqual(
            len(facet_state_filtered_response.data["results"]), self.published_count
        )

        # Should contain 2 facets
        self.assertEqual(len(facet_state_filtered_response.data["facets"]), 2)
        # With 3 statuses
        self.assertEqual(
            len(
                facet_state_filtered_response.data["facets"]["_filter_state_global"][
                    "state_global"
                ]["buckets"]
            ),
            3,
        )

        # Still same facets
        self.assertIn(
            {"doc_count": self.published_count, "key": "published"},
            facet_state_response.data["facets"]["_filter_state_global"]["state_global"][
                "buckets"
            ],
        )
        self.assertIn(
            {"doc_count": self.in_progress_count, "key": "in_progress"},
            facet_state_response.data["facets"]["_filter_state_global"]["state_global"][
                "buckets"
            ],
        )
        self.assertIn(
            {"doc_count": self.rejected_count, "key": "rejected"},
            facet_state_response.data["facets"]["_filter_state_global"]["state_global"][
                "buckets"
            ],
        )

    def test_list_results_with_facets(self):
        """Test list results with facets."""
        return self._list_results_with_facets()


if __name__ == "__main__":
    unittest.main()



"""
Suggesters backend.

It's assumed, that fields you're planning to query suggestions for have been
properly indexed using ``fields.CompletionField``.

Example:

    >>> from django_elasticsearch_dsl import Document, Index, fields
    >>>
    >>> from books.models import Publisher
    >>>
    >>> # Name of the Elasticsearch index
    >>> PUBLISHER_INDEX = Index(PUBLISHER_INDEX_NAME)
    >>> # See Elasticsearch Indices API reference for available settings
    >>> PUBLISHER_INDEX.settings(
    >>>     number_of_shards=1,
    >>>     number_of_replicas=1
    >>> )
    >>>
    >>> @PUBLISHER_INDEX.doc_type
    >>> class PublisherDocument(Document):
    >>>     "Publisher Elasticsearch document."
    >>>
    >>>     id = fields.IntegerField(attr='id')
    >>>
    >>>     name = fields.StringField(
    >>>         fields={
    >>>             'raw': fields.StringField(analyzer='keyword'),
    >>>             'suggest': fields.CompletionField(),
    >>>         }
    >>>     )
    >>>
    >>>     info = fields.StringField()
    >>>
    >>>     address = fields.StringField(
    >>>         fields={
    >>>             'raw': fields.StringField(analyzer='keyword')
    >>>         }
    >>>     )
    >>>
    >>>     city = fields.StringField(
    >>>         fields={
    >>>             'raw': fields.StringField(analyzer='keyword'),
    >>>             'suggest': fields.CompletionField(),
    >>>         }
    >>>     )
    >>>
    >>>     state_province = fields.StringField(
    >>>         fields={
    >>>             'raw': fields.StringField(analyzer='keyword'),
    >>>             'suggest': fields.CompletionField(),
    >>>         }
    >>>     )
    >>>
    >>>     country = fields.StringField(
    >>>         fields={
    >>>             'raw': fields.StringField(analyzer='keyword'),
    >>>             'suggest': fields.CompletionField(),
    >>>         }
    >>>     )
    >>>
    >>>     website = fields.StringField()
    >>>
    >>>     class Meta:
    >>>         "Meta options."
    >>>
    >>>         model = Publisher  # The model associate with this Document
"""

from collections import defaultdict

from django_elasticsearch_dsl_drf_alt.constants import (
    SUGGESTER_TERM,
    SUGGESTER_PHRASE,
    SUGGESTER_COMPLETION,
    ALL_SUGGESTERS,
)

from rest_framework.filters import BaseFilterBackend

from six import string_types

from ..mixins import FilterBackendMixin

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.suggester"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("SuggesterFilterBackend",)


class SuggesterFilterBackend(BaseFilterBackend, FilterBackendMixin):
    """Suggester filter backend for Elasticsearch.

    Suggestion functionality is exclusive. Once you have queried the
    ``SuggesterFilterBackend``, the latter will transform your current
    search query into suggestion search query (which is very different).
    Therefore, always add it as the very last filter backend.

    Example:

        >>> from django_elasticsearch_dsl_drf_alt.constants import (
        >>>     SUGGESTER_TERM,
        >>>     SUGGESTER_PHRASE,
        >>>     SUGGESTER_COMPLETION,
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.filter_backends import (
        >>>     SuggesterFilterBackend
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.viewsets import DocumentViewSet
        >>>
        >>> # Local PublisherDocument definition
        >>> from .documents import PublisherDocument
        >>>
        >>> # Local PublisherDocument serializer
        >>> from .serializers import PublisherDocumentSerializer
        >>>
        >>> class PublisherDocumentView(BaseDocumentViewSet):
        >>>
        >>>     document = PublisherDocument
        >>>     serializer_class = PublisherDocumentSerializer
        >>>     filter_backends = [
        >>>         # ...
        >>>         SuggesterFilterBackend,
        >>>     ]
        >>>     # Suggester fields
        >>>     suggester_fields = {
        >>>         'name_suggest': {
        >>>             'field': 'name.suggest',
        >>>             'suggesters': [
        >>>                 SUGGESTER_TERM,
        >>>                 SUGGESTER_PHRASE,
        >>>                 SUGGESTER_COMPLETION,
        >>>             ],
        >>>         },
        >>>         'city_suggest': {
        >>>             'field': 'city.suggest',
        >>>             'suggesters': [
        >>>                 SUGGESTER_COMPLETION,
        >>>             ],
        >>>         },
        >>>         'state_province_suggest': {
        >>>             'field': 'state_province.suggest',
        >>>             'suggesters': [
        >>>                 SUGGESTER_COMPLETION,
        >>>             ],
        >>>         },
        >>>         'country_suggest': {
        >>>             'field': 'country.suggest',
        >>>             'suggesters': [
        >>>                 SUGGESTER_COMPLETION,
        >>>             ],
        >>>         },
        >>>     }
    """

    @classmethod
    def prepare_suggester_fields(cls, view):
        """Prepare filter fields.

        :param view:
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Filtering options.
        :rtype: dict
        """
        filter_fields = view.suggester_fields

        for field, options in filter_fields.items():
            if options is None or isinstance(options, string_types):
                filter_fields[field] = {"field": options or field}
            elif "field" not in filter_fields[field]:
                filter_fields[field]["field"] = field

            if "suggesters" not in filter_fields[field]:
                filter_fields[field]["suggesters"] = tuple(ALL_SUGGESTERS)

        return filter_fields

    @classmethod
    def get_suggester_context(cls, field, suggester_name, request, view):
        """Get suggester context.

        Given the following definition (in ViewSets):

            >>> # Suggester fields
            >>> suggester_fields = {
            >>>     'title_suggest': {
            >>>         'field': 'title.suggest',
            >>>         'default_suggester': SUGGESTER_COMPLETION,
            >>>     },
            >>>     'title_suggest_context': {
            >>>         'field': 'title.suggest_context',
            >>>         'default_suggester': SUGGESTER_COMPLETION,
            >>>         'completion_options': {
            >>>             'filters': {
            >>>                 'title_suggest_tag': 'tag',
            >>>                 'title_suggest_state': 'state',
            >>>                 'title_suggest_publisher': 'publisher',
            >>>             },
            >>>             'size': 10,
            >>>         }
            >>>     },
            >>> }

        http://localhost:8000/search/books-frontend/suggest/?title_suggest_context=M

        When talking about the queries made, we have the following. Multiple
        values per field are combined with OR:

            >>> completion={
            >>>     'field': options['field'],
            >>>     'size': 10,
            >>>     'contexts': {
            >>>         'tag': ['History', 'Drama'],
            >>>     }
            >>> }

        The following works with OR as well, so it seems we have OR only.
        However, the following construction is more handy, since it allows
        us to play with boosting nicely. Also, it allows to provide `prefix`
        param (which in case of the example given below means that suggestions
        shall match both categories with "Child", "Children", "Childrend's").
        Simply put it's treated as `prefix`, rather than `term`.

            >>> completion={
            >>>     'field': options['field'],
            >>>     'size': 10,
            >>>     'contexts': {
            >>>         'tag': [
            >>>             {'context': 'History'},
            >>>             {'context': 'Drama', 'boost': 2},
            >>>             {'context': 'Children', 'prefix': True},
            >>>         ],
            >>>     },
            >>> }

        Sample query for `category` filter:

            /search/books-frontend/suggest/
            ?title_suggest_context=M
            &title_suggest_tag=Art__2.0
            &title_suggest_tag=Documentary__2.0__prefix
            &title_suggest_publisher=Apress

        The query params would be:

            query_params:
            <QueryDict: {
              'title_suggest_context': ['M'],
              'title_suggest_tag': ['Art__2.0', 'Documentary__2.0__prefix'],
              'title_suggest_publisher': ['Apress']
            }>

        Sample query for `geo` filter:

            /search/address/suggest/
            ?street_suggest_context=M
            &street_suggest_loc=43.66__-79.22__2.0__10000km

        The query params would be:

            query_params:
            <QueryDict: {
              'street_suggest_context': ['M'],
              'street_suggest_loc': ['Art__43.66__-79.22__2.0__10000km'],
            }>

        :return:
        """
        contexts = {}
        query_params = request.query_params.copy()

        # Processing `category` filters:
        for query_param, context_field in (
            field["completion_options"].get("category_filters", {}).items()
        ):
            context_field_query = defaultdict(list)
            for context_field_value in query_params.getlist(query_param, []):
                context_field_value_parts = cls.split_lookup_filter(
                    context_field_value, maxsplit=2
                )
                # Case `&title_suggest_tag=Documentary__2.0__prefix`
                if len(context_field_value_parts) == 3:
                    context_field_query[context_field].append(
                        {
                            "context": context_field_value_parts[0],
                            "boost": context_field_value_parts[1],
                            "prefix": True,
                        }
                    )
                # Case `&title_suggest_tag=Documentary__2.0` or
                # `&title_suggest_tag=Documentary__prefix`.
                elif len(context_field_value_parts) == 2:
                    if context_field_value_parts[1] == "prefix":
                        context_field_query[context_field].append(
                            {
                                "context": context_field_value_parts[0],
                                "prefix": True,
                            }
                        )
                    else:
                        context_field_query[context_field].append(
                            {
                                "context": context_field_value_parts[0],
                                "boost": context_field_value_parts[1],
                            }
                        )
                # Case `&title_suggest_tag=Documentary`.
                elif len(context_field_value_parts) == 1:
                    context_field_query[context_field].append(
                        {
                            "context": context_field_value_parts[0],
                        }
                    )

            contexts.update(context_field_query)

        # Processing `geo` filters:
        for query_param, context_field in (
            field["completion_options"].get("geo_filters", {}).items()
        ):
            context_field_query = defaultdict(list)
            for context_field_value in query_params.getlist(query_param, []):
                context_field_value_parts = cls.split_lookup_filter(
                    context_field_value, maxsplit=3
                )
                # Case `&street_suggest_loc=43.66__-79.22__2.0__10000km`
                if len(context_field_value_parts) == 4:
                    context_field_query[context_field].append(
                        {
                            "context": {
                                "lat": context_field_value_parts[0],
                                "lon": context_field_value_parts[1],
                            },
                            "precision": context_field_value_parts[2],
                            "boost": context_field_value_parts[3],
                        }
                    )
                # Case `&street_suggest_loc=43.66__-79.22__10000km`.
                elif len(context_field_value_parts) == 3:
                    context_field_query[context_field].append(
                        {
                            "context": {
                                "lat": context_field_value_parts[0],
                                "lon": context_field_value_parts[1],
                            },
                            "precision": context_field_value_parts[2],
                        }
                    )
                # Case `&street_suggest_loc=43.66__-79.22`.
                elif len(context_field_value_parts) == 2:
                    context_field_query[context_field].append(
                        {
                            "context": {
                                "lat": context_field_value_parts[0],
                                "lon": context_field_value_parts[1],
                            },
                        }
                    )

            contexts.update(context_field_query)

        return contexts

    @classmethod
    def apply_suggester_term(cls, suggester_name, queryset, options, value):
        """Apply `term` suggester.

        :param suggester_name:
        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type suggester_name: str
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return queryset.suggest(suggester_name, value, term={"field": options["field"]})

    @classmethod
    def apply_suggester_phrase(cls, suggester_name, queryset, options, value):
        """Apply `phrase` suggester.

        :param suggester_name:
        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type suggester_name: str
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return queryset.suggest(
            suggester_name, value, phrase={"field": options["field"]}
        )

    @classmethod
    def apply_suggester_completion(cls, suggester_name, queryset, options, value):
        """Apply `completion` suggester.

        :param suggester_name:
        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type suggester_name: str
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        completion_kwargs = {
            "field": options["field"],
        }
        if "size" in options:
            completion_kwargs["size"] = options["size"]
        if "contexts" in options:
            completion_kwargs["contexts"] = options["contexts"]
        if "skip_duplicates" in options:
            completion_kwargs["skip_duplicates"] = options["skip_duplicates"]
        return queryset.suggest(suggester_name, value, completion=completion_kwargs)

    def get_suggester_query_params(self, request, view):
        """Get query params to be for suggestions.

        :param request: Django REST framework request.
        :param view: View.
        :type request: rest_framework.request.Request
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Request query params to filter on.
        :rtype: dict
        """
        query_params = request.query_params.copy()

        suggester_query_params = {}
        suggester_fields = self.prepare_suggester_fields(view)
        for query_param in query_params:
            query_param_list = self.split_lookup_filter(query_param, maxsplit=1)
            field_name = query_param_list[0]

            if field_name in suggester_fields:
                suggester_param = None
                if len(query_param_list) > 1:
                    suggester_param = query_param_list[1]

                valid_suggesters = suggester_fields[field_name]["suggesters"]

                # If we have default suggester given use it as a default and
                # do not require further suffix specification.
                default_suggester = None
                if "default_suggester" in suggester_fields[field_name]:
                    default_suggester = suggester_fields[field_name][
                        "default_suggester"
                    ]

                if suggester_param is None or suggester_param in valid_suggesters:

                    # If we have default suggester given use it as a default
                    # and do not require further suffix specification.
                    if suggester_param is None and default_suggester is not None:
                        suggester_param = str(default_suggester)

                    values = [
                        __value.strip()
                        for __value in query_params.getlist(query_param)
                        if __value.strip() != ""
                    ]

                    if values:
                        _sf = suggester_fields[field_name]
                        suggester_query_params[query_param] = {
                            "suggester": suggester_param,
                            "values": values,
                            "field": suggester_fields[field_name].get(
                                "field", field_name
                            ),
                            "type": view.mapping,
                        }

                        if "options" in _sf:
                            if "size" in _sf["options"]:
                                suggester_query_params[query_param].update(
                                    {"size": _sf["options"]["size"]}
                                )
                            if "skip_duplicates" in _sf["options"]:
                                suggester_query_params[query_param].update(
                                    {
                                        "skip_duplicates": _sf["options"][
                                            "skip_duplicates"
                                        ]
                                    }
                                )

                        if (
                            suggester_param == SUGGESTER_COMPLETION
                            and "completion_options" in _sf
                            and (
                                "category_filters" in _sf["completion_options"]
                                or "geo_filters" in _sf["completion_options"]
                            )
                        ):
                            suggester_query_params[query_param]["contexts"] = (
                                self.get_suggester_context(
                                    suggester_fields[field_name],
                                    suggester_param,
                                    request,
                                    view,
                                )
                            )

        return suggester_query_params

    def filter_queryset(self, request, queryset, view):
        """Filter the queryset.

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        # The ``SuggesterFilterBackend`` filter backend shall be used in
        # the ``suggest`` custom view action/route only. Usages outside of the
        # are ``suggest`` action/route are restricted.
        if view.action != "suggest":
            return queryset

        suggester_query_params = self.get_suggester_query_params(request, view)
        for suggester_name, options in suggester_query_params.items():
            # We don't have multiple values here.
            for value in options["values"]:
                # `term` suggester
                if options["suggester"] == SUGGESTER_TERM:
                    queryset = self.apply_suggester_term(
                        suggester_name, queryset, options, value
                    )

                # `phrase` suggester
                elif options["suggester"] == SUGGESTER_PHRASE:
                    queryset = self.apply_suggester_phrase(
                        suggester_name, queryset, options, value
                    )

                # `completion` suggester
                elif options["suggester"] == SUGGESTER_COMPLETION:
                    queryset = self.apply_suggester_completion(
                        suggester_name, queryset, options, value
                    )

        return queryset



"""
Test filtering backend.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

from books import constants
import factories
from search_indexes.viewsets import BookDocumentViewSet

from ..constants import (
    SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE,
    SEPARATOR_LOOKUP_COMPLEX_VALUE,
    SEPARATOR_LOOKUP_FILTER,
    SEPARATOR_LOOKUP_NAME,
)
from ..filter_backends import FilteringFilterBackend
from .base import (
    BaseRestFrameworkTestCase,
    CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED,
    CORE_API_AND_CORE_SCHEMA_MISSING_MSG,
)
from .data_mixins import AddressesMixin, BooksMixin

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_filtering_common"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestFilteringCommon",)


@pytest.mark.django_db
class TestFilteringCommon(BaseRestFrameworkTestCase, AddressesMixin, BooksMixin):
    """Test filtering common."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up."""
        super(TestFilteringCommon, cls).setUpClass()
        # Testing simple documents: Publisher index.
        cls.create_books()

        # Testing nested objects: Addresses, cities and countries
        cls.created_addresses()

        cls.sleep()
        # Update the Elasticsearch index
        call_command("search_index", "--rebuild", "-f")

        # Testing coreapi and coreschema
        cls.backend = FilteringFilterBackend()
        cls.view = BookDocumentViewSet()

        cls.books_default_filter_lookup_url = reverse(
            "bookdocument_default_filter_lookup-list", kwargs={}
        )

    # ***********************************************************************
    # ************************ Simple fields ********************************
    # ***********************************************************************

    def _field_filter_value(self, field_name, value, count):
        """Field filter value.

        Usage example:

            >>> self._field_filter_value(
            >>>     'title__wildcard',
            >>>     self.prefix[3:-3],
            >>>     self.prefix_count
            >>> )
        """
        url = self.base_url[:]
        data = {}
        response = self.client.get(url + "?{}={}".format(field_name, value), data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), count)

    def _field_filter_multiple_values(self, url, field_name, values, count):
        """Field filter multiple values.

        Usage example:

            >>> self._field_filter_multiple_values(
            >>>     self.books_default_filter_lookup_url,
            >>>     'authors',
            >>>     ['Author 1', 'Author 2'],
            >>>     3
            >>> )
        """
        # url = self.base_url[:]
        data = {}
        params = "&".join(["{}={}".format(field_name, __v) for __v in values])
        response = self.client.get(url + "?{}".format(params), data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), count)

    def _field_filter_term(self, field_name, filter_value):
        """Field filter term.

        Example:

            http://localhost:8000/api/articles/?tags=children
        """
        self.authenticate()

        url = self.base_url[:]
        data = {}

        # Should contain 22 results
        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), self.all_count)

        # Should contain only 10 results
        filtered_response = self.client.get(
            url + "?{}={}".format(field_name, filter_value), data
        )
        self.assertEqual(filtered_response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(filtered_response.data["results"]), self.published_count)

    def test_field_filter_term(self):
        """Field filter term."""
        return self._field_filter_term(
            "state", constants.BOOK_PUBLISHING_STATUS_PUBLISHED
        )

    def test_field_filter_term_explicit(self):
        """Field filter term."""
        return self._field_filter_term(
            "state__term", constants.BOOK_PUBLISHING_STATUS_PUBLISHED
        )

    def test_field_filter_range(self):
        """Field filter range.

        Example:

            http://localhost:8000/api/users/?age__range=16__67
        """
        lower_id = self.published[0].id
        upper_id = self.published[-1].id
        value = "{lower_id}{separator}{upper_id}".format(
            lower_id=lower_id,
            separator=SEPARATOR_LOOKUP_COMPLEX_VALUE,
            upper_id=upper_id,
        )
        return self._field_filter_value("id__range", value, self.published_count)

    def test_field_filter_regexp(self):
        """Field filter regexp.

        Example:

            http://localhost:8000/api/users/?title__regexp=De.{8}Ins.*
        """
        return self._field_filter_value(
            "title__regexp",
            "{}.{{8}}{}.*".format(self.prefix[0:2], self.prefix[10:13]),
            self.prefix_count,
        )

    def test_field_filter_range_with_boost(self):
        """Field filter range.

        Example:

            http://localhost:8000/api/users/?age__range=16__67__2.0
        """
        lower_id = self.published[0].id
        upper_id = self.published[-1].id
        value = "{lower_id}{separator}{upper_id}{separator}{boost}".format(
            lower_id=lower_id,
            upper_id=upper_id,
            boost="2.0",
            separator=SEPARATOR_LOOKUP_COMPLEX_VALUE,
        )
        return self._field_filter_value("id__range", value, self.published_count)

    def test_field_filter_prefix(self):
        """Test filter prefix.

        Example:

            http://localhost:8000/api/articles/?tags__prefix=bio
        """
        return self._field_filter_value("title__prefix", self.prefix, self.prefix_count)

    def test_field_filter_in(self):
        """Test filter in.

        Example:

            http://localhost:8000/api/articles/?id__in=1__2__3
        """
        return self._field_filter_value(
            "id__in",
            SEPARATOR_LOOKUP_COMPLEX_VALUE.join([str(__b.id) for __b in self.prefixed]),
            self.prefix_count,
        )

    def _field_filter_terms_list(self, field_name, in_values, count):
        """Field filter terms.

        Example:

            http://localhost:8000/api/articles/?id=1&id=2&id=3
        """
        url = self.base_url[:]
        data = {}
        url_parts = ["{}={}".format(field_name, val) for val in in_values]
        response = self.client.get(url + "?{}".format("&".join(url_parts)), data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), count)

    def test_field_filter_terms_list(self):
        """Test filter terms."""
        return self._field_filter_terms_list(
            "id", [str(__b.id) for __b in self.prefixed], self.prefix_count
        )

    def test_field_filter_terms_string(self):
        """Test filter terms.

        Example:

            http://localhost:8000/api/articles/?id__terms=1__2__3
        """
        return self._field_filter_value(
            "id__terms",
            SEPARATOR_LOOKUP_COMPLEX_VALUE.join([str(__b.id) for __b in self.prefixed]),
            self.prefix_count,
        )

    def test_field_filter_exists_true(self):
        """Test filter exists true.

        Example:

            http://localhost:8000/api/articles/?tags__exists=true
        """
        return self._field_filter_value("tags__exists", "true", self.all_count)

    def test_field_filter_exists_false(self):
        """Test filter exists.

        Example:

            http://localhost:8000/api/articles/?non_existent__exists=false
        """
        return self._field_filter_value(
            "non_existent_field__exists", "false", self.all_count
        )

    def test_field_filter_wildcard(self):
        """Test filter wildcard.

        Example:

            http://localhost:8000/api/articles/?title__wildcard=*elusional*
        """
        return self._field_filter_value(
            "title__wildcard", "*{}*".format(self.prefix[1:6]), self.prefix_count
        )

    def test_field_filter_exclude(self):
        """Test filter exclude.

        Example:

            http://localhost:8000/api/articles/?tags__exclude=children
        """
        return self._field_filter_value(
            "state__exclude",
            constants.BOOK_PUBLISHING_STATUS_PUBLISHED,
            self.all_count - self.published_count,
        )

    def test_field_filter_isnull_true(self):
        """Test filter isnull true.

        Example:

            http://localhost:8000/api/articles/?null_field__isnull=true
        """
        self._field_filter_value("null_field__isnull", "true", self.all_count)
        self._field_filter_value("tags__isnull", "true", self.no_tags_count)

    def test_field_filter_isnull_false(self):
        """Test filter isnull true.

        Example:

            http://localhost:8000/api/articles/?tags__isnull=false
        """
        self._field_filter_value("null_field__isnull", "false", 0)
        self._field_filter_value(
            "tags__isnull", "false", self.all_count - self.no_tags_count
        )

    def test_field_filter_endswith(self):
        """Test filter endswith.

        Example:

            http://localhost:8000/api/articles/?state__endswith=lished
        """
        return self._field_filter_value(
            "state__endswith",
            constants.BOOK_PUBLISHING_STATUS_PUBLISHED[4:],
            self.published_count,
        )

    def test_field_filter_contains(self):
        """Test filter contains.

        Example:

            http://localhost:8000/api/articles/?state__contains=lishe
        """
        return self._field_filter_value(
            "state__contains",
            constants.BOOK_PUBLISHING_STATUS_PUBLISHED[4:-2],
            self.published_count,
        )

    def _field_filter_gte_lte(self, field_name, value, lookup, boost=None):
        """Field filter gt/gte/lt/lte.

        Example:

            http://localhost:8000/api/users/?id__gt=10
            http://localhost:8000/api/users/?id__gte=10
            http://localhost:8000/api/users/?id__lt=10
            http://localhost:8000/api/users/?id__lte=10
        """
        url = self.base_url[:]
        data = {}

        if boost is not None:
            url += "?{field_name}__{lookup}={value}{separator}{boost}".format(
                field_name=field_name,
                lookup=lookup,
                value=value,
                boost=boost,
                separator=SEPARATOR_LOOKUP_COMPLEX_VALUE,
            )
        else:
            url += "?{field_name}__{lookup}={value}".format(
                field_name=field_name, lookup=lookup, value=value
            )

        response = self.client.get(url, data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        __mapping = {
            "gt": self.assertGreater,
            "gte": self.assertGreaterEqual,
            "lt": self.assertLess,
            "lte": self.assertLessEqual,
        }
        __func = __mapping.get(lookup)

        if callable(__func):
            for obj in response.data["results"]:
                __func(obj["id"], value)

    def test_field_filter_gt(self):
        """Field filter gt.

        Example:

            http://localhost:8000/api/users/?id__gt=10
        :return:
        """
        return self._field_filter_gte_lte("id", self.in_progress[0].id, "gt")

    def test_field_filter_gt_with_boost(self):
        """Field filter gt with boost.

        Example:

            http://localhost:8000/api/users/?id__gt=10__2.0
        :return:
        """
        # TODO: check boost value
        return self._field_filter_gte_lte("id", self.in_progress[0].id, "gt", "2.0")

    def test_field_filter_gte(self):
        """Field filter gte.

        Example:

            http://localhost:8000/api/users/?id__gte=10
        :return:
        """
        return self._field_filter_gte_lte("id", self.in_progress[0].id, "gte")

    def test_field_filter_lt(self):
        """Field filter lt.

        Example:

            http://localhost:8000/api/users/?id__lt=10
        :return:
        """
        return self._field_filter_gte_lte("id", self.in_progress[0].id, "lt")

    def test_field_filter_lt_with_boost(self):
        """Field filter lt with boost.

        Example:

            http://localhost:8000/api/users/?id__lt=10__2.0
        :return:
        """
        # TODO: check boost value
        return self._field_filter_gte_lte("id", self.in_progress[0].id, "lt", "2.0")

    def test_field_filter_lte(self):
        """Field filter lte.

        Example:

            http://localhost:8000/api/users/?id__lte=10
        :return:
        """
        return self._field_filter_gte_lte("id", self.in_progress[0].id, "lte")

    def test_ids_filter(self):
        """Test ids filter.

        Example:

            http://localhost:8000/api/articles/?ids=68__64__58
            http://localhost:8000/api/articles/?ids=68&ids=64&ids=58
        """
        __ids = [str(__obj.id) for __obj in self.published]
        return self._field_filter_value(
            "ids", SEPARATOR_LOOKUP_COMPLEX_VALUE.join(__ids), self.published_count
        )

    def test_ids_empty_filter(self):
        """Test ids filter with empty value. This should not fail.

        Example:

            http://localhost:8000/api/articles/?ids=
        """
        __ids = []
        return self._field_filter_value(
            "ids", SEPARATOR_LOOKUP_COMPLEX_VALUE.join(__ids), 0
        )

    def test_default_filter_lookup(self):
        """Test default filter lookup.

        Example:

            http://localhost:8000/search/books-default-filter-lookup/
                ?authors=Robin&authors=Luc
        """
        # Create two authors
        author_1 = factories.AuthorFactory(name="Author1")
        author_2 = factories.AuthorFactory(name="Author2")
        authors = [author_1, author_2]
        # Add them to 3 books
        self.published[0].authors.add(*authors)
        self.published[1].authors.add(*authors)
        self.published[2].authors.add(*authors)
        # Update the Elasticsearch index
        call_command("search_index", "--rebuild", "-f")
        # Test
        self._field_filter_multiple_values(
            self.books_default_filter_lookup_url, "authors", authors, 3
        )

    # ***********************************************************************
    # ************************ Nested fields ********************************
    # ***********************************************************************

    def _nested_field_filter_term(self, field_name, filter_value, count):
        """Nested field filter term.

        Example:

            http://localhost:8000/api/articles/?tags=children
        """
        self.authenticate()

        url = self.addresses_url[:]
        data = {}

        # Should contain only 32 results
        response = self.client.get(url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), self.all_addresses_count)

        # Should contain only 10 results
        filtered_response = self.client.get(
            url + "?{}={}".format(field_name, filter_value), data
        )
        self.assertEqual(filtered_response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(filtered_response.data["results"]), count)

    def test_nested_field_filter_term(self):
        """Nested field filter term."""
        self._nested_field_filter_term(
            "city", "Yerevan", self.addresses_in_yerevan_count
        )

        self._nested_field_filter_term(
            "country", "Armenia", self.addresses_in_yerevan_count
        )

        self._nested_field_filter_term("city", "Dublin", self.addresses_in_dublin_count)

    # ***********************************************************************
    # ************************* Other fields ********************************
    # ***********************************************************************

    def test_various_complex_fields(self):
        """Test various complex fields.

        :return:
        """
        data = {}
        response = self.client.get(self.cities_url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response = self.client.get(self.city_detail_url, data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    # ***********************************************************************
    # ******************** Core api and core schema *************************
    # ***********************************************************************

    @unittest.skipIf(
        not CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED, CORE_API_AND_CORE_SCHEMA_MISSING_MSG
    )
    def test_schema_fields_with_filter_fields_list(self):
        """Test schema field generator"""
        fields = self.backend.get_schema_fields(self.view)
        fields = [f.name for f in fields]
        self.assertEqual(fields, list(self.view.filter_fields.keys()))

    @unittest.skipIf(
        not CORE_API_AND_CORE_SCHEMA_ARE_INSTALLED, CORE_API_AND_CORE_SCHEMA_MISSING_MSG
    )
    def test_schema_field_not_required(self):
        """Test schema fields always not required"""
        fields = self.backend.get_schema_fields(self.view)
        fields = [f.required for f in fields]
        for field in fields:
            self.assertFalse(field)


if __name__ == "__main__":
    unittest.main()



"""
Search filter backends.
"""

from .base import BaseSearchFilterBackend
from .compound import CompoundSearchFilterBackend
from .historical import SearchFilterBackend
from .multi_match import MultiMatchSearchFilterBackend
from .simple_query_string import SimpleQueryStringSearchFilterBackend

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.search"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "BaseSearchFilterBackend",
    "CompoundSearchFilterBackend",
    "MultiMatchSearchFilterBackend",
    "SearchFilterBackend",
    "SimpleQueryStringSearchFilterBackend",
)



from django.core.management.base import BaseCommand

from ...elasticsearch_helpers import delete_all_indices, get_all_indices


class Command(BaseCommand):
    help = 'Remove all indexes from Elasticsearch'

    def add_arguments(self, parser):
        # Named (optional) arguments
        parser.add_argument(
            '--dry-run',
            action='store_true',
            dest='dry_run',
            default=False,
            help='Dry-run (no db writes)',
        )
        parser.add_argument(
            '--with-protected',
            action='store_true',
            dest='with_protected',
            default=False,
            help='Including protected (for instance, kibana) indexes',
        )

    def handle(self, *args, **options):
        dry_run = options.get('dry_run', False)
        with_protected = options.get('with_protected', False)
        if dry_run:
            indices = get_all_indices(with_protected=with_protected)
            print("The following indexes will be removed: {}".format(indices))
        else:
            indices, errors = delete_all_indices(with_protected=with_protected)
            print("The following indexes are removed: {}".format(indices))
            print(
                "The following indexes could not be removed: {}".format(errors)
            )



"""
Common filtering backend.
"""

import operator

from elasticsearch_dsl.query import Q
from rest_framework.filters import BaseFilterBackend
from django_elasticsearch_dsl import fields

import six
from six import string_types

from ...constants import (
    TRUE_VALUES,
    FALSE_VALUES,
    ALL_LOOKUP_FILTERS_AND_QUERIES,
    LOOKUP_FILTER_PREFIX,
    LOOKUP_FILTER_RANGE,
    LOOKUP_FILTER_REGEXP,
    LOOKUP_FILTER_TERMS,
    LOOKUP_FILTER_EXISTS,
    LOOKUP_FILTER_WILDCARD,
    LOOKUP_QUERY_CONTAINS,
    LOOKUP_QUERY_IN,
    LOOKUP_QUERY_GT,
    LOOKUP_QUERY_GTE,
    LOOKUP_QUERY_LT,
    LOOKUP_QUERY_LTE,
    LOOKUP_QUERY_STARTSWITH,
    LOOKUP_QUERY_ENDSWITH,
    LOOKUP_QUERY_ISNULL,
    LOOKUP_QUERY_EXCLUDE,
)
from ..mixins import FilterBackendMixin

from ...compat import coreapi, coreschema


__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.filtering.common"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("FilteringFilterBackend",)


class FilteringFilterBackend(BaseFilterBackend, FilterBackendMixin):
    """Filtering filter backend for Elasticsearch.

    Example:

        >>> from django_elasticsearch_dsl_drf_alt.constants import (
        >>>     LOOKUP_FILTER_PREFIX,
        >>>     LOOKUP_FILTER_WILDCARD,
        >>>     LOOKUP_QUERY_EXCLUDE,
        >>>     LOOKUP_QUERY_ISNULL,
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.filter_backends import (
        >>>     FilteringFilterBackend
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.viewsets import (
        >>>     BaseDocumentViewSet,
        >>> )
        >>>
        >>> # Local article document definition
        >>> from .documents import ArticleDocument
        >>>
        >>> # Local article document serializer
        >>> from .serializers import ArticleDocumentSerializer
        >>>
        >>> class ArticleDocumentView(BaseDocumentViewSet):
        >>>
        >>>     document = ArticleDocument
        >>>     serializer_class = ArticleDocumentSerializer
        >>>     filter_backends = [FilteringFilterBackend,]
        >>>     filter_fields = {
        >>>         'title': 'title.raw',
        >>>         'state': {
        >>>             'field': 'state.raw',
        >>>             'lookups': [
        >>>                 LOOKUP_FILTER_PREFIX,
        >>>                 LOOKUP_FILTER_WILDCARD,
        >>>                 LOOKUP_QUERY_EXCLUDE,
        >>>                 LOOKUP_QUERY_ISNULL,
        >>>             ],
        >>>             'default_lookup': LOOKUP_FILTER_WILDCARD,
        >>>         }
        >>> }
    """

    @classmethod
    def prepare_filter_fields(cls, view):
        """Prepare filter fields.

        :param view:
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Filtering options.
        :rtype: dict
        """
        filter_fields = view.filter_fields

        for field, options in filter_fields.items():
            if options is None or isinstance(options, string_types):
                filter_fields[field] = {"field": options or field}
            elif "field" not in filter_fields[field]:
                filter_fields[field]["field"] = field

            if "lookups" not in filter_fields[field]:
                filter_fields[field]["lookups"] = tuple(ALL_LOOKUP_FILTERS_AND_QUERIES)

        return filter_fields

    @classmethod
    def get_range_params(cls, value):
        """Get params for `range` query.

        Syntax:

            /endpoint/?field_name__range={lower}__{upper}__{boost}
            /endpoint/?field_name__range={lower}__{upper}

        Example:

            http://localhost:8000/api/users/?age__range=16__67__2.0
            http://localhost:8000/api/users/?age__range=16__67
            http://localhost:8000/api/users/?age__range=16

        :param value:
        :type: str
        :return: Params to be used in `range` query.
        :rtype: dict
        """
        __values = cls.split_lookup_complex_value(value, maxsplit=3)
        __len_values = len(__values)

        if __len_values == 0:
            return {}

        params = {"gte": __values[0]}

        if __len_values == 3:
            params["lte"] = __values[1]
            params["boost"] = __values[2]
            params.update(
                {
                    "lte": __values[1],
                    "boost": __values[2],
                }
            )
        elif __len_values == 2:
            params["lte"] = __values[1]

        return params

    @classmethod
    def get_gte_lte_params(cls, value, lookup):
        """Get params for `gte`, `gt`, `lte` and `lt` query.

        Syntax:

            /endpoint/?field_name__gt={lower}__{boost}
            /endpoint/?field_name__gt={lower}

        Example:

            http://localhost:8000/api/articles/?id__gt=1
            http://localhost:8000/api/articles/?id__gt=1__2.0

        :param value:
        :param lookup:
        :type value: str
        :type lookup: str
        :return: Params to be used in `range` query.
        :rtype: dict
        """
        __values = cls.split_lookup_complex_value(value, maxsplit=2)
        __len_values = len(__values)

        if __len_values == 0:
            return {}

        params = {lookup: __values[0]}

        if __len_values == 2:
            params["boost"] = __values[1]

        return params

    @classmethod
    def apply_filter_term(cls, queryset, options, value):
        """Apply `term` filter.

        Syntax:

            /endpoint/?field_name={value}

        Example:

            http://localhost:8000/api/articles/?tags=children

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["term"],
            kwargs={options["field"]: value},
        )

    @classmethod
    def apply_filter_terms(cls, queryset, options, value):
        """Apply `terms` filter.

        Syntax:

            /endpoint/?field_name__terms={value1}__{value2}
            /endpoint/?field_name__terms={value1}

        Note, that number of values is not limited.

        Example:

            http://localhost:8000/api/articles/?tags__terms=children__python
            http://localhost:8000/api/articles/?tags__terms=children

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: mixed: either str or iterable (list, tuple).
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        # If value is a list or a tuple, we use it as is.
        if isinstance(value, (list, tuple)):
            __values = value

        # Otherwise, we consider it to be a string and split it further.
        else:
            __values = cls.split_lookup_complex_value(value)

        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["terms"],
            kwargs={options["field"]: __values},
        )

    @classmethod
    def apply_filter_range(cls, queryset, options, value):
        """Apply `range` filter.

         Syntax:

            /endpoint/?field_name__range={lower}__{upper}__{boost}
            /endpoint/?field_name__range={lower}__{upper}

        Example:

            http://localhost:8000/api/users/?age__range=16__67__2.0
            http://localhost:8000/api/users/?age__range=16__67
            http://localhost:8000/api/users/?age__range=16

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["range"],
            kwargs={options["field"]: cls.get_range_params(value)},
        )

    @classmethod
    def apply_filter_regexp(cls, queryset, options, value):
        """Apply `reexp` filter.

         Syntax:

            /endpoint/?field_name__regexp={regexp}

        Example:

            http://localhost:8000/api/users/?age__regexp=1[6-9]
            http://localhost:8000/api/users/?age__regexp=2.*

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["regexp"],
            kwargs={options["field"]: value},
        )

    @classmethod
    def apply_query_exists(cls, queryset, options, value):
        """Apply `exists` filter.

        Syntax:

            /endpoint/?field_name__exists=true
            /endpoint/?field_name__exists=false

        Example:

            http://localhost:8000/api/articles/?tags__exists=true
            http://localhost:8000/api/articles/?tags__exists=false

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        _value_lower = value.lower()
        if _value_lower in TRUE_VALUES:
            return cls.apply_query(
                queryset=queryset,
                options=options,
                args=[Q("exists", field=options["field"])],
            )
        elif _value_lower in FALSE_VALUES:
            return cls.apply_query(
                queryset=queryset,
                options=options,
                args=[~Q("exists", field=options["field"])],
            )
        return queryset

    @classmethod
    def apply_filter_prefix(cls, queryset, options, value):
        """Apply `prefix` filter.

        Syntax:

            /endpoint/?field_name__prefix={value}

        Example:

            http://localhost:8000/api/articles/?tags__prefix=bio

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["prefix"],
            kwargs={options["field"]: value},
        )

    @classmethod
    def apply_query_wildcard(cls, queryset, options, value):
        """Apply `wildcard` filter.

        Syntax:

            /endpoint/?field_name__wildcard={value}*
            /endpoint/?field_name__wildcard=*{value}
            /endpoint/?field_name__wildcard=*{value}*

        Example:

            http://localhost:8000/api/articles/?tags__wildcard=child*

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_query(
            queryset=queryset,
            options=options,
            args=[Q("wildcard", **{options["field"]: value})],
        )

    @classmethod
    def apply_query_contains(cls, queryset, options, value):
        """Apply `contains` filter.

        Syntax:

            /endpoint/?field_name__contains={value}

        Example:

            http://localhost:8000/api/articles/?state__contains=lis

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_query(
            queryset=queryset,
            options=options,
            args=[Q("wildcard", **{options["field"]: "*{}*".format(value)})],
        )

    @classmethod
    def apply_query_endswith(cls, queryset, options, value):
        """Apply `endswith` filter.

        Syntax:

            /endpoint/?field_name__endswith={value}

        Example:

            http://localhost:8000/api/articles/?tags__endswith=dren

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_query(
            queryset=queryset,
            options=options,
            args=[Q("wildcard", **{options["field"]: "*{}".format(value)})],
        )

    @classmethod
    def apply_query_in(cls, queryset, options, value):
        """Apply `in` functional query.

        Syntax:

            /endpoint/?field_name__in={value1}__{value2}
            /endpoint/?field_name__in={value1}

        Note, that number of values is not limited.

        Example:

            http://localhost:8000/api/articles/?tags__in=children__python

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        __values = cls.split_lookup_complex_value(value)
        __queries = []
        for __value in __values:
            __queries.append(Q("term", **{options["field"]: __value}))

        if __queries:
            queryset = cls.apply_query(
                queryset=queryset,
                options=options,
                args=[six.moves.reduce(operator.or_, __queries)],
            )

        return queryset

    @classmethod
    def apply_query_gt(cls, queryset, options, value):
        """Apply `gt` functional query.

        Syntax:

            /endpoint/?field_name__gt={value}__{boost}
            /endpoint/?field_name__gt={value}

        Example:

            http://localhost:8000/api/articles/?id__gt=1__2.0
            http://localhost:8000/api/articles/?id__gt=1

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["range"],
            kwargs={options["field"]: cls.get_gte_lte_params(value, "gt")},
        )

    @classmethod
    def apply_query_gte(cls, queryset, options, value):
        """Apply `gte` functional query.

        Syntax:

            /endpoint/?field_name__gte={value}__{boost}
            /endpoint/?field_name__gte={value}

        Example:

            http://localhost:8000/api/articles/?id__gte=1__2.0
            http://localhost:8000/api/articles/?id__gte=1

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["range"],
            kwargs={options["field"]: cls.get_gte_lte_params(value, "gte")},
        )

    @classmethod
    def apply_query_lt(cls, queryset, options, value):
        """Apply `lt` functional query.

        Syntax:

            /endpoint/?field_name__lt={value}__{boost}
            /endpoint/?field_name__lt={value}

        Example:

            http://localhost:8000/api/articles/?id__lt=1__2.0
            http://localhost:8000/api/articles/?id__lt=1

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["range"],
            kwargs={options["field"]: cls.get_gte_lte_params(value, "lt")},
        )

    @classmethod
    def apply_query_lte(cls, queryset, options, value):
        """Apply `lte` functional query.

        Syntax:

            /endpoint/?field_name__lte={value}__{boost}
            /endpoint/?field_name__lte={value}

        Example:

            http://localhost:8000/api/articles/?id__lte=1__2.0
            http://localhost:8000/api/articles/?id__lte=1

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["range"],
            kwargs={options["field"]: cls.get_gte_lte_params(value, "lte")},
        )

    @classmethod
    def apply_query_isnull(cls, queryset, options, value):
        """Apply `isnull` functional query.

        Syntax:

            /endpoint/?field_name__isnull=true
            /endpoint/?field_name__isnull=false

        Example:

            http://localhost:8000/api/articles/?tags__isnull=true
            http://localhost:8000/api/articles/?tags__isnull=false

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        _value_lower = value.lower()
        if _value_lower in TRUE_VALUES:
            return cls.apply_query(
                queryset=queryset,
                options=options,
                args=[~Q("exists", field=options["field"])],
            )
        elif _value_lower in FALSE_VALUES:
            return cls.apply_query(
                queryset=queryset,
                options=options,
                args=[Q("exists", field=options["field"])],
            )
        return queryset

    @classmethod
    def apply_query_exclude(cls, queryset, options, value):
        """Apply `exclude` functional query.

        Syntax:

            /endpoint/?field_name__isnull={value1}__{value2}
            /endpoint/?field_name__exclude={valu1}

        Note, that number of values is not limited.

        Example:

            http://localhost:8000/api/articles/?tags__exclude=children__python
            http://localhost:8000/api/articles/?tags__exclude=children

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        __values = cls.split_lookup_complex_value(value)

        __queries = []
        for __value in __values:
            __queries.append(~Q("term", **{options["field"]: __value}))

        if __queries:
            queryset = cls.apply_query(
                queryset=queryset,
                options=options,
                args=[six.moves.reduce(operator.or_, __queries)],
            )

        return queryset

    def get_filter_query_params(self, request, view):
        """Get query params to be filtered on.

        :param request: Django REST framework request.
        :param view: View.
        :type request: rest_framework.request.Request
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Request query params to filter on.
        :rtype: dict
        """
        query_params = request.query_params.copy()

        filter_query_params = {}
        filter_fields = self.prepare_filter_fields(view)
        for query_param in query_params:
            query_param_list = self.split_lookup_filter(query_param, maxsplit=1)
            field_name = query_param_list[0]

            if field_name in filter_fields:
                lookup_param = None
                if len(query_param_list) > 1:
                    lookup_param = query_param_list[1]

                valid_lookups = filter_fields[field_name]["lookups"]

                # If we have default lookup given use it as a default and
                # do not require further suffix specification.
                default_lookup = None
                if "default_lookup" in filter_fields[field_name]:
                    default_lookup = filter_fields[field_name]["default_lookup"]

                if lookup_param is None or lookup_param in valid_lookups:

                    # If we have default lookup given use it as a default
                    # and do not require further suffix specification.
                    if lookup_param is None and default_lookup is not None:
                        lookup_param = str(default_lookup)

                    values = [
                        __value.strip()
                        for __value in query_params.getlist(query_param)
                        if __value.strip() != ""
                    ]

                    if values:
                        filter_query_params[query_param] = {
                            "lookup": lookup_param,
                            "values": values,
                            "field": filter_fields[field_name].get("field", field_name),
                            "type": view.mapping,
                        }
        return filter_query_params

    def filter_queryset(self, request, queryset, view):
        """Filter the queryset.

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        filter_query_params = self.get_filter_query_params(request, view)
        for options in filter_query_params.values():
            # When no specific lookup given, in case of multiple values
            # we apply `terms` filter by default and proceed to the next
            # query param.
            if (
                isinstance(options["values"], (list, tuple))
                and options["lookup"] is None
            ):
                queryset = self.apply_filter_terms(queryset, options, options["values"])
                continue

            # For all other cases, when we don't have multiple values,
            # we follow the normal flow.
            for value in options["values"]:
                # `terms` filter lookup
                if options["lookup"] == LOOKUP_FILTER_TERMS:
                    queryset = self.apply_filter_terms(queryset, options, value)

                # `prefix` filter lookup
                elif options["lookup"] in (
                    LOOKUP_FILTER_PREFIX,
                    LOOKUP_QUERY_STARTSWITH,
                ):
                    queryset = self.apply_filter_prefix(queryset, options, value)

                # `range` filter lookup
                elif options["lookup"] == LOOKUP_FILTER_RANGE:
                    queryset = self.apply_filter_range(queryset, options, value)

                # `regexp` filter lookup
                elif options["lookup"] == LOOKUP_FILTER_REGEXP:
                    queryset = self.apply_filter_regexp(queryset, options, value)

                # `exists` filter lookup
                elif options["lookup"] == LOOKUP_FILTER_EXISTS:
                    queryset = self.apply_query_exists(queryset, options, value)

                # `wildcard` filter lookup
                elif options["lookup"] == LOOKUP_FILTER_WILDCARD:
                    queryset = self.apply_query_wildcard(queryset, options, value)

                # `contains` filter lookup
                elif options["lookup"] == LOOKUP_QUERY_CONTAINS:
                    queryset = self.apply_query_contains(queryset, options, value)

                # `in` functional query lookup
                elif options["lookup"] == LOOKUP_QUERY_IN:
                    queryset = self.apply_query_in(queryset, options, value)

                # `gt` functional query lookup
                elif options["lookup"] == LOOKUP_QUERY_GT:
                    queryset = self.apply_query_gt(queryset, options, value)

                # `gte` functional query lookup
                elif options["lookup"] == LOOKUP_QUERY_GTE:
                    queryset = self.apply_query_gte(queryset, options, value)

                # `lt` functional query lookup
                elif options["lookup"] == LOOKUP_QUERY_LT:
                    queryset = self.apply_query_lt(queryset, options, value)

                # `lte` functional query lookup
                elif options["lookup"] == LOOKUP_QUERY_LTE:
                    queryset = self.apply_query_lte(queryset, options, value)

                # `endswith` filter lookup
                elif options["lookup"] == LOOKUP_QUERY_ENDSWITH:
                    queryset = self.apply_query_endswith(queryset, options, value)

                # `isnull` functional query lookup
                elif options["lookup"] == LOOKUP_QUERY_ISNULL:
                    queryset = self.apply_query_isnull(queryset, options, value)

                # `exclude` functional query lookup
                elif options["lookup"] == LOOKUP_QUERY_EXCLUDE:
                    queryset = self.apply_query_exclude(queryset, options, value)

                # `term` filter lookup. This is default if no `default_lookup`
                # option has been given or explicit lookup provided.
                else:
                    queryset = self.apply_filter_term(queryset, options, value)

        return queryset

    def get_coreschema_field(self, field):
        if isinstance(field, fields.IntegerField):
            field_cls = coreschema.Number
        else:
            field_cls = coreschema.String
        return field_cls()

    def get_schema_fields(self, view):
        assert coreapi is not None, (
            "coreapi must be installed to " "use `get_schema_fields()`"
        )
        assert coreschema is not None, (
            "coreschema must be installed to " "use `get_schema_fields()`"
        )
        filter_fields = getattr(view, "filter_fields", None)
        document = getattr(view, "document", None)

        return (
            []
            if not filter_fields
            else [
                coreapi.Field(
                    name=field_name,
                    required=False,
                    location="query",
                    schema=self.get_coreschema_field(document._fields.get(field_name)),
                )
                for field_name in filter_fields
            ]
        )



# coding: utf-8
"""
Test more-like-this functionality.
"""

from __future__ import absolute_import, unicode_literals

import unittest
import logging

from django.core.management import call_command
from django.urls import reverse

import pytest

from rest_framework import status

import factories

from .base import BaseRestFrameworkTestCase

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_suggesters"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestMoreLikeThis",)

LOGGER = logging.getLogger(__name__)


@pytest.mark.django_db
class TestMoreLikeThis(BaseRestFrameworkTestCase):
    """Test suggesters."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up class."""
        super(TestMoreLikeThis, cls).setUpClass()

        cls.lorem_books = factories.BookFactory.create_batch(200)

        cls.books = []

        cls.alice_books = []
        for book_data in factories.constants.NON_FAKER_BOOK_CONTENT:
            cls.alice_books.append(
                factories.BookChapterFactory(
                    title=book_data["title"],
                    summary=book_data["summary"],
                    description=book_data["description"],
                )
            )
        cls.alice_books_ids = [__obj.id for __obj in cls.alice_books]

        cls.sheckley_books = []
        for book_data in factories.constants.NON_FAKER_BOOK_CONTENT_OTHER:
            cls.sheckley_books.append(
                factories.BookNovelFactory(
                    title=book_data["title"],
                    summary=book_data["summary"],
                    description=book_data["description"],
                )
            )
        cls.sheckley_books_ids = [__obj.id for __obj in cls.sheckley_books]

        cls.books = []
        cls.books.extend(cls.alice_books)
        cls.books.extend(cls.sheckley_books)

        # Alice book
        cls.books_url_1 = reverse(
            "bookdocument_more_like_this-more-like-this", kwargs={"id": cls.books[0].id}
        )
        # Sheckley book
        cls.books_url_2 = reverse(
            "bookdocument_more_like_this-more-like-this",
            kwargs={"id": cls.books[-1].id},
        )

        # Alice book
        cls.books_url_1_no_options = reverse(
            "bookdocument_more_like_this_no_options-more-like-this",
            kwargs={"id": cls.books[0].id},
        )
        # Sheckley book
        cls.books_url_2_no_options = reverse(
            "bookdocument_more_like_this_no_options-more-like-this",
            kwargs={"id": cls.books[-1].id},
        )

        cls.sleep()
        call_command("search_index", "--rebuild", "-f")

    def _test_more_like_this(self, test_data_ids, url, strict=True):
        """Test more-like-this.

        We can't really predict which result would it show as most relevant,
        however we can for sure assume that no strange data shall appear
        in between.
        """
        self.authenticate()
        data = {}
        response = self.client.get(url, data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        if strict:
            response_ids = [int(__r["id"]) for __r in response.data["results"]]

            self.assertFalse(bool(set(response_ids) - set(test_data_ids)))

    def test_more_like_this(self):
        """Test more-like-this."""
        # Testing publishers
        test_data_ids = []
        test_data_ids.extend(self.alice_books_ids)
        test_data_ids.extend(self.sheckley_books_ids)

        self._test_more_like_this(test_data_ids, self.books_url_1, strict=True)
        self._test_more_like_this(
            test_data_ids, self.books_url_1_no_options, strict=False
        )

        self._test_more_like_this(test_data_ids, self.books_url_2, strict=True)
        self._test_more_like_this(
            test_data_ids, self.books_url_2_no_options, strict=False
        )


if __name__ == "__main__":
    unittest.main()



"""
Test pagination.
"""

from __future__ import absolute_import

import unittest

from django.core.management import call_command
from django.urls import reverse

from elasticsearch.connection.base import Connection

import pytest

from rest_framework import status

import factories

from .base import BaseRestFrameworkTestCase

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_query_friendly_pagination"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("TestQueryFriendlyPagination",)

old_log_request_success = Connection.log_request_success
es_call_count = 0


def patched_log_request_success(self, *args, **kwargs):
    global es_call_count
    es_call_count += 1
    old_log_request_success(self, *args, **kwargs)


Connection.log_request_success = patched_log_request_success


@pytest.mark.django_db
class TestQueryFriendlyPagination(BaseRestFrameworkTestCase):
    """Test pagination."""

    pytestmark = pytest.mark.django_db

    @classmethod
    def setUpClass(cls):
        """Set up class."""
        super(TestQueryFriendlyPagination, cls).setUpClass()
        cls.publishers = factories.PublisherFactory.create_batch(43)
        cls.books = factories.BookFactory.create_batch(43)

        cls.sleep()
        call_command("search_index", "--rebuild", "-f")

    def _test_pagination(self):
        """Test pagination."""
        invalid_page_url = self.books_url + "?page=3&page_size=30"

        invalid_response = self.client.get(invalid_page_url, self.data)
        self.assertEqual(invalid_response.status_code, status.HTTP_404_NOT_FOUND)

    def _test_pagination_orphans(self):
        """Test pagination returning orphaned nodes"""
        valid_page_url = self.books_url + "?page=1&page_size=40&orphans=3"

        # Check if response now is valid
        valid_response = self.client.get(valid_page_url, self.data)
        self.assertEqual(valid_response.status_code, status.HTTP_200_OK)

        # Check totals
        self.assertEqual(len(valid_response.data["results"]), 43)

    def _test_pagination_orphans_over(self):
        """Test pagination when orphaned nodes fall into next page"""
        valid_page_url = self.books_url + "?page=1&page_size=40&orphans=2"

        # Check if response now is valid
        valid_response = self.client.get(valid_page_url, self.data)
        self.assertEqual(valid_response.status_code, status.HTTP_200_OK)

        # Check totals
        self.assertEqual(len(valid_response.data["results"]), 40)

        valid_page_url = self.books_url + "?page=2&page_size=40&orphans=2"

        valid_response = self.client.get(valid_page_url, self.data)
        self.assertEqual(valid_response.status_code, status.HTTP_200_OK)

        # Check totals
        self.assertEqual(len(valid_response.data["results"]), 3)

    def _test_pagination_offset(self):
        """Test pagination defined by offset and limit"""
        valid_page_url = self.publishers_url + "?limit=5&offset=8"
        valid_response = self.client.get(valid_page_url, self.data)
        self.assertEqual(valid_response.status_code, status.HTTP_200_OK)

        # Check totals
        self.assertEqual(len(valid_response.data["results"]), 5)

    def test_pagination(self):
        """Test pagination."""
        self.authenticate()
        self.publishers_url = reverse("publisherdocument-list", kwargs={})
        self.books_url = reverse(
            "bookdocument_query_friendly_pagination-list", kwargs={}
        )
        self.data = {}

        last_es_call_count = es_call_count
        self._test_pagination()
        # Ensure number of ES calls is only 1
        self.assertEqual(es_call_count - last_es_call_count, 1)

        last_es_call_count = es_call_count
        self._test_pagination_orphans()
        # Only in case of orphaned nodes, we need 1 more fallback ES call
        self.assertEqual(es_call_count - last_es_call_count, 2)

        last_es_call_count = es_call_count
        self._test_pagination_orphans_over()
        # Here are two requests
        self.assertEqual(es_call_count - last_es_call_count, 2)

        last_es_call_count = es_call_count
        self._test_pagination_offset()
        self.assertEqual(es_call_count - last_es_call_count, 1)


if __name__ == "__main__":
    unittest.main()



import json

__title__ = "django_elasticsearch_dsl_drf_alt.wrappers"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = (
    "dict_to_obj",
    "obj_to_dict",
    "Wrapper",
)


class Wrapper(object):
    """Wrapper.

    Example:
    >>> from django_elasticsearch_dsl_drf_alt.wrappers import dict_to_obj
    >>>
    >>> mapping = {
    >>>     'country': {
    >>>         'name': 'Netherlands',
    >>>         'province': {
    >>>             'name': 'North Holland',
    >>>             'city': {
    >>>                 'name': 'Amsterdam',
    >>>             }
    >>>         }
    >>>     }
    >>> }
    >>>
    >>> wrapper = dict_to_obj(mapping)
    >>> wrapper.country.name
    >>> "Netherlands"
    >>> wrapper.country.province.name
    >>> "North Holland"
    >>> wrapper.country.province.city.name
    >>> "Amsterdam"
    >>> wrapper.as_dict
    >>> {
    >>>     'country': {
    >>>         'name': 'Netherlands',
    >>>         'province': {
    >>>             'name': 'North Holland',
    >>>             'city': {
    >>>                 'name': 'Amsterdam',
    >>>             }
    >>>         }
    >>>     }
    >>> }
    >>> str(wrapper)
    >>> "Netherlands"
    """

    def __str__(self):
        for key, item in self.__dict__.items():
            if isinstance(item, Wrapper):
                return item.__str__()
            else:
                return item

    @property
    def as_dict(self):
        """As dict.

        :return:
        :rtype: dict
        """
        return obj_to_dict(self)

    @property
    def as_json(self):
        """As JSON.

        :return:
        :rtype: str
        """
        return json.dumps(self.as_dict)


def dict_to_obj(mapping):
    """dict to obj mapping.

    :param mapping:
    :type mapping: dict
    :return:
    :rtype: :obj:`Wrapper`
    """
    wrapper = Wrapper()

    for key, item in mapping.items():
        if isinstance(item, dict):
            setattr(wrapper, key, dict_to_obj(item))
        else:
            setattr(wrapper, key, item)
    return wrapper


def obj_to_dict(obj):
    """Wrapper to dict.

    :param obj:
    :type obj: `obj`:Wrapper:
    :return:
    :rtype: dict
    """
    mapping = {}

    for key, item in obj.__dict__.items():
        if isinstance(item, Wrapper):
            mapping.update({key: obj_to_dict(item)})
        else:
            mapping.update({key: item})

    return mapping



import copy

from elasticsearch_dsl.query import Q

from .base import BaseSearchQueryBackend

__title__ = (
    "django_elasticsearch_dsl_drf_alt.filter_backends.search."
    "query_backends.simple_query_string"
)
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("SimpleQueryStringQueryBackend",)


class SimpleQueryStringQueryBackend(BaseSearchQueryBackend):
    """Simple query string query backend."""

    query_type = "simple_query_string"

    @classmethod
    def get_field(cls, field, options):
        """Get field.

        :param field:
        :param options:
        :return:
        """
        if not options:
            options = {}

        field_name = options["field"] if "field" in options else field

        if "boost" in options:
            return "{}^{}".format(field_name, options["boost"])
        return field_name

    @classmethod
    def get_query_options(cls, request, view, search_backend):
        query_options = getattr(view, "simple_query_string_options", {})
        return query_options

    @classmethod
    def construct_search(cls, request, view, search_backend):
        """Construct search.

        In case of multi match, we always look in a group of fields.
        Thus, matching per field is no longer valid use case here. However,
        we might want to have multiple fields enabled for multi match per
        view set, and only search in some of them in specific request.

        Example:

            /search/books/?search_simple_query_string=
                "fried eggs" %2B(eggplant | potato) -frittata
            /search/books/?search_simple_query_string=
                title,summary:"fried eggs" +(eggplant | potato) -frittata

        Note, that multiple searches are not supported (would not raise
        an exception, but would simply take only the first):

            /search/books/?search_simple_query_string=
                title,summary:"fried eggs" +(eggplant | potato) -frittata
                &search_simple_query_string=
                author,publisher="fried eggs" +(eggplant | potato) -frittata

        In the view-set fields shall be defined in a very simple way. The
        only accepted argument would be boost (per field).

        Example 1 (complex):

            simple_query_string_search_fields = {
                'title': {'field': 'title.english', 'boost': 4},
                'summary': {'boost': 2},
                'description': None,
            }

        Example 2 (simple list):

            simple_query_string_search_fields = (
                'title',
                'summary',
                'description',
            )

        Query examples:

            http://localhost:8000/search
                /books-simple-query-string-search-backend
                /?search_simple_query_string=%22Pool%20of%20Tears%22

            http://localhost:8000/search
                /books-simple-query-string-search-backend
                /?search_simple_query_string=%22Pool%20of%20Tears%22
                -considering

        :param request:
        :param view:
        :param search_backend:
        :return:
        """
        if hasattr(view, "simple_query_string_search_fields"):
            view_search_fields = copy.copy(
                getattr(view, "simple_query_string_search_fields")
            )
        else:
            view_search_fields = copy.copy(view.search_fields)

        __is_complex = isinstance(view_search_fields, dict)

        # Getting the list of search query params.
        query_params = search_backend.get_search_query_params(request)

        __queries = []
        for search_term in query_params[:1]:
            __values = search_backend.split_lookup_name(search_term, 1)
            __len_values = len(__values)
            __search_term = search_term

            query_fields = []

            # If we're dealing with case like
            # /search/books/?search_multi_match=title,summary:lorem ipsum
            if __len_values > 1:
                _field, value = __values
                __search_term = value
                fields = search_backend.split_lookup_complex_multiple_value(_field)
                for field in fields:
                    if field in view_search_fields:
                        if __is_complex:
                            query_fields.append(
                                cls.get_field(field, view_search_fields[field])
                            )
                        else:
                            query_fields.append(field)

            # If it's just a simple search like
            # /search/books/?search_multi_match=lorem ipsum
            # Fields shall be defined in a very simple way.
            else:
                # It's a dict, see example 1 (complex)
                if __is_complex:
                    for field, options in view_search_fields.items():
                        query_fields.append(cls.get_field(field, options))

                # It's a list, see example 2 (simple)
                else:
                    query_fields = copy.copy(view_search_fields)

            # The multi match query
            __queries.append(
                Q(
                    cls.query_type,
                    query=__search_term,
                    fields=query_fields,
                    **cls.get_query_options(request, view, search_backend)
                )
            )

        return __queries



import unittest
import mock

# For Python3 >= 3.4
try:
    from importlib import reload
# For Python3 < 3.4
except ImportError as err:
    try:
        from imp import reload
    except ImportError as err:
        pass

__title__ = "django_elasticsearch_dsl_drf_alt.tests.test_versions"
__author__ = "Artur Barseghyan"
__copyright__ = "Copyright (c) 2017-2020 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("VersionsTest",)


class VersionsTest(unittest.TestCase):
    """
    Tests of ``django_elasticsearch_dsl_drf_alt.versions`` module.
    """

    def setUp(self):
        pass

    @mock.patch("elasticsearch_dsl.__version__", [6, 3, 0])
    def test_elasticsearch_dsl_6_3_0(self):
        """
        Tests as if we were using elasticsearch_dsl==6.3.0.
        """
        from django_elasticsearch_dsl_drf import versions

        reload(versions)

        # Exact version matching
        self.assertFalse(versions.ELASTICSEARCH_7_0)
        self.assertTrue(versions.ELASTICSEARCH_6_3)
        self.assertFalse(versions.ELASTICSEARCH_6_2)
        self.assertFalse(versions.ELASTICSEARCH_5_4)
        self.assertFalse(versions.ELASTICSEARCH_2_0)
        self.assertFalse(versions.ELASTICSEARCH_6_2)

        # Less than or equal matching
        self.assertFalse(versions.ELASTICSEARCH_LTE_2_0)
        self.assertFalse(versions.ELASTICSEARCH_LTE_2_2)
        self.assertFalse(versions.ELASTICSEARCH_LTE_5_0)
        self.assertFalse(versions.ELASTICSEARCH_LTE_5_3)
        self.assertFalse(versions.ELASTICSEARCH_LTE_6_0)
        self.assertTrue(versions.ELASTICSEARCH_LTE_6_3)
        self.assertTrue(versions.ELASTICSEARCH_LTE_7_0)
        self.assertTrue(versions.ELASTICSEARCH_LTE_8_0)

        # Greater than or equal matching
        self.assertTrue(versions.ELASTICSEARCH_GTE_2_0)
        self.assertTrue(versions.ELASTICSEARCH_GTE_2_2)
        self.assertTrue(versions.ELASTICSEARCH_GTE_5_0)
        self.assertTrue(versions.ELASTICSEARCH_GTE_5_3)
        self.assertTrue(versions.ELASTICSEARCH_GTE_6_0)
        self.assertTrue(versions.ELASTICSEARCH_GTE_6_3)
        self.assertFalse(versions.ELASTICSEARCH_GTE_7_0)
        self.assertFalse(versions.ELASTICSEARCH_GTE_8_0)

    @mock.patch("elasticsearch_dsl.__version__", [7, 0, 0])
    def test_elasticsearch_dsl_7_0_0(self):
        """
        Tests as if we were using elasticsearch_dsl==7.0.0.
        """
        from django_elasticsearch_dsl_drf import versions

        reload(versions)

        # Exact version matching
        self.assertTrue(versions.ELASTICSEARCH_7_0)
        self.assertFalse(versions.ELASTICSEARCH_6_3)
        self.assertFalse(versions.ELASTICSEARCH_6_2)
        self.assertFalse(versions.ELASTICSEARCH_5_4)
        self.assertFalse(versions.ELASTICSEARCH_2_0)
        self.assertFalse(versions.ELASTICSEARCH_6_2)

        # Less than or equal matching
        self.assertFalse(versions.ELASTICSEARCH_LTE_2_0)
        self.assertFalse(versions.ELASTICSEARCH_LTE_2_2)
        self.assertFalse(versions.ELASTICSEARCH_LTE_5_0)
        self.assertFalse(versions.ELASTICSEARCH_LTE_5_3)
        self.assertFalse(versions.ELASTICSEARCH_LTE_6_0)
        self.assertFalse(versions.ELASTICSEARCH_LTE_6_3)
        self.assertTrue(versions.ELASTICSEARCH_LTE_7_0)
        self.assertTrue(versions.ELASTICSEARCH_LTE_8_0)

        # Greater than or equal matching
        self.assertTrue(versions.ELASTICSEARCH_GTE_2_0)
        self.assertTrue(versions.ELASTICSEARCH_GTE_2_2)
        self.assertTrue(versions.ELASTICSEARCH_GTE_5_0)
        self.assertTrue(versions.ELASTICSEARCH_GTE_5_3)
        self.assertTrue(versions.ELASTICSEARCH_GTE_6_0)
        self.assertTrue(versions.ELASTICSEARCH_GTE_6_3)
        self.assertTrue(versions.ELASTICSEARCH_GTE_7_0)
        self.assertFalse(versions.ELASTICSEARCH_GTE_8_0)


if __name__ == "__main__":
    unittest.main()



"""
Geo spatial filtering backend.

Elasticsearch supports two types of geo data:

- geo_point fields which support lat/lon pairs
- geo_shape fields, which support points, lines, circles, polygons,
  multi-polygons etc.

The queries in this group are:

- geo_shape query: Find document with geo-shapes which either intersect,
  are contained by, or do not intersect with the specified geo-shape.
- geo_bounding_box query: Finds documents with geo-points that fall into
  the specified rectangle.
+ geo_distance query: Finds document with geo-points within the specified
  distance of a central point.
- geo_distance_range query: Like the geo_distance query, but the range
  starts at a specified distance from the central point. Note, that this
  one is deprecated and this isn't implemented.
+ geo_polygon query: Find documents with geo-points within the specified
  polygon.
"""

import logging
from elasticsearch_dsl.query import Q
from rest_framework.filters import BaseFilterBackend

from six import string_types

from ...constants import (
    ALL_GEO_SPATIAL_LOOKUP_FILTERS_AND_QUERIES,
    LOOKUP_FILTER_GEO_DISTANCE,
    LOOKUP_FILTER_GEO_POLYGON,
    LOOKUP_FILTER_GEO_BOUNDING_BOX,
    LOOKUP_FILTER_GEO_SHAPE,
    SEPARATOR_LOOKUP_COMPLEX_VALUE,
    SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE,
)
from ..mixins import FilterBackendMixin

__title__ = "django_elasticsearch_dsl_drf_alt.filter_backends.filtering.common"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("GeoSpatialFilteringFilterBackend",)


LOGGER = logging.getLogger(__name__)


class GeoSpatialFilteringFilterBackend(BaseFilterBackend, FilterBackendMixin):
    """Geo-spatial filtering filter backend for Elasticsearch.

    Example:

        >>> from django_elasticsearch_dsl_drf_alt.constants import (
        >>>     LOOKUP_FILTER_GEO_DISTANCE,
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.filter_backends import (
        >>>     GeoSpatialFilteringFilterBackend
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.viewsets import (
        >>>     BaseDocumentViewSet,
        >>> )
        >>>
        >>> # Local article document definition
        >>> from .documents import ArticleDocument
        >>>
        >>> # Local article document serializer
        >>> from .serializers import ArticleDocumentSerializer
        >>>
        >>> class ArticleDocumentView(BaseDocumentViewSet):
        >>>
        >>>     document = ArticleDocument
        >>>     serializer_class = ArticleDocumentSerializer
        >>>     filter_backends = [GeoSpatialFilteringFilterBackend,]
        >>>     geo_spatial_filter_fields = {
        >>>         'loc': 'location',
        >>>         'location': {
        >>>             'field': 'location',
        >>>             'lookups': [
        >>>                 LOOKUP_FILTER_GEO_DISTANCE,
        >>>             ],
        >>>         }
        >>> }
    """

    @classmethod
    def prepare_filter_fields(cls, view):
        """Prepare filter fields.

        :param view:
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Filtering options.
        :rtype: dict
        """
        filter_fields = view.geo_spatial_filter_fields

        for field, options in filter_fields.items():
            if options is None or isinstance(options, string_types):
                filter_fields[field] = {"field": options or field}
            elif "field" not in filter_fields[field]:
                filter_fields[field]["field"] = field

            if "lookups" not in filter_fields[field]:
                filter_fields[field]["lookups"] = tuple(
                    ALL_GEO_SPATIAL_LOOKUP_FILTERS_AND_QUERIES
                )

        return filter_fields

    @classmethod
    def get_geo_distance_params(cls, value, field):
        """Get params for `geo_distance` query.

        Example:

            /api/articles/?location__geo_distance=2km__43.53__-12.23

        :param value:
        :param field:
        :type value: str
        :type field:
        :return: Params to be used in `geo_distance` query.
        :rtype: dict
        """
        __values = cls.split_lookup_complex_value(value, maxsplit=3)
        __len_values = len(__values)

        if __len_values < 3:
            return {}

        params = {
            "distance": __values[0],
            field: {
                "lat": __values[1],
                "lon": __values[2],
            },
        }

        if __len_values == 4:
            params["distance_type"] = __values[3]
        else:
            params["distance_type"] = "arc"

        return params

    @classmethod
    def get_geo_polygon_params(cls, value, field):
        """Get params for `geo_polygon` query.

        Example:

            /api/articles/?location__geo_polygon=40,-70__30,-80__20,-90

        Example:

            /api/articles/?location__geo_polygon=40,-70
                __30,-80
                __20,-90
                ___name,myname
                __validation_method,IGNORE_MALFORMED

        Elasticsearch:

            {
                "query": {
                    "bool" : {
                        "must" : {
                            "match_all" : {}
                        },
                        "filter" : {
                            "geo_polygon" : {
                                "person.location" : {
                                    "points" : [
                                        {"lat" : 40, "lon" : -70},
                                        {"lat" : 30, "lon" : -80},
                                        {"lat" : 20, "lon" : -90}
                                    ]
                                }
                            }
                        }
                    }
                }
            }

        :param value:
        :param field:
        :type value: str
        :type field:
        :return: Params to be used in `geo_distance` query.
        :rtype: dict
        """
        __values = cls.split_lookup_complex_value(value)
        __len_values = len(__values)

        if not __len_values:
            return {}

        __points = []
        __options = {}

        for __value in __values:
            if SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE in __value:
                __split_value = __value.split(SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE)
                if len(__split_value) >= 2:
                    if __split_value[0] in ("_name", "validation_method"):
                        __options.update({__split_value[0]: __split_value[1]})
                    else:
                        __points.append(
                            {
                                "lat": float(__split_value[0]),
                                "lon": float(__split_value[1]),
                            }
                        )

        if __points:
            params = {field: {"points": __points}}
            params.update(__options)

            return params
        return {}

    @classmethod
    def get_geo_bounding_box_params(cls, value, field):
        """Get params for `geo_bounding_box` query.

        Example:

            /api/articles/?location__geo_bounding_box=40.73,-74.1__40.01,-71.12

        Example:

            /api/articles/?location__geo_polygon=40.73,-74.1
                __40.01,-71.12
                ___name,myname
                __validation_method,IGNORE_MALFORMED
                __type,indexed

        Elasticsearch:

            {
                "query": {
                    "bool" : {
                        "must" : {
                            "match_all" : {}
                        },
                        "filter" : {
                            "geo_bounding_box" : {
                                "person.location" : {
                                    "top_left" : {
                                        "lat" : 40.73,
                                        "lon" : -74.1
                                    },
                                    "bottom_right" : {
                                        "lat" : 40.01,
                                        "lon" : -71.12
                                    }
                                }
                            }
                        }
                    }
                }
            }

        :param value:
        :param field:
        :type value: str
        :type field:
        :return: Params to be used in `geo_bounding_box` query.
        :rtype: dict
        """
        __values = cls.split_lookup_complex_value(value)
        __len_values = len(__values)

        if not __len_values:
            return {}

        __top_left_points = {}
        __bottom_right_points = {}
        __options = {}

        # Top left
        __lat_lon = __values[0].split(SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE)
        if len(__lat_lon) >= 2:
            __top_left_points.update(
                {
                    "lat": float(__lat_lon[0]),
                    "lon": float(__lat_lon[1]),
                }
            )

        # Bottom right
        __lat_lon = __values[1].split(SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE)
        if len(__lat_lon) >= 2:
            __bottom_right_points.update(
                {
                    "lat": float(__lat_lon[0]),
                    "lon": float(__lat_lon[1]),
                }
            )

        # Options
        for __value in __values[2:]:
            if SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE in __value:
                __opt_name_val = __value.split(SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE)
                if len(__opt_name_val) >= 2:
                    if __opt_name_val[0] in ("_name", "validation_method", "type"):
                        __options.update({__opt_name_val[0]: __opt_name_val[1]})

        if not __top_left_points or not __bottom_right_points:
            return {}

        params = {
            field: {
                "top_left": __top_left_points,
                "bottom_right": __bottom_right_points,
            }
        }
        params.update(__options)

        return params

    @classmethod
    def get_geo_shape_params(cls, value, field):
        """Get params for `geo_shape` query.

        Example:

            /search/publishers/?location__geo_shape=48.9864453,6.37977
                __relation,intersects
                __type,circle
                __radius,20km

        Example:

            /search/publishers/?location__geo_shape=48.906254,6.378593
                __48.985850,6.479359
                __relation,within
                __type,envelope

        Elasticsearch:

            {
                "query": {
                    "bool" : {
                        "must" : {
                            "match_all" : {}
                        },
                        "filter" : {
                            "geo_shape" : {
                                "location" : {
                                    "shape": {
                                        "type": "circle",
                                        "coordinates": [48.9864453, 6.37977],
                                        "radius": "20km"
                                    },
                                    "relation": "intersects"
                                }
                            }
                        }
                    }
                }
            }

        :param value:
        :param field:
        :type value: str
        :type field:
        :return: Params to be used in `geo_shape` query.
        :rtype: dict
        """
        __values = cls.split_lookup_complex_value(value)
        __len_values = len(__values)

        if not __len_values:
            return {}

        __coordinates = []
        __options = {}

        # Parse coordinates (can be x points)
        for value in __values:
            __lat_lon = value.split(SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE)
            if len(__lat_lon) >= 2:
                try:
                    __point = [
                        float(__lat_lon[0]),
                        float(__lat_lon[1]),
                    ]
                    __coordinates.append(list(__point))
                except ValueError:
                    if SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE in value:
                        __opt_name_val = value.split(
                            SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE
                        )
                        if len(__opt_name_val) >= 2:
                            if __opt_name_val[0] in ("relation", "type", "radius"):
                                __options.update({__opt_name_val[0]: __opt_name_val[1]})

        __type = __options.pop("type", None)
        __relation = __options.pop("relation", None)
        if not __coordinates or not __type or not __relation:
            return {}

        params = {
            field: {
                "shape": {
                    "type": __type,
                    "coordinates": (
                        __coordinates if len(__coordinates) > 1 else __coordinates[0]
                    ),
                },
                "relation": __relation,
            }
        }
        radius = __options.pop("radius", None)
        if radius:
            params[field]["shape"].update({"radius": radius})
        params.update(__options)

        return params

    @classmethod
    def apply_query_geo_distance(cls, queryset, options, value):
        """Apply `geo_distance` query.

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return queryset.query(
            Q("geo_distance", **cls.get_geo_distance_params(value, options["field"]))
        )

    @classmethod
    def apply_query_geo_polygon(cls, queryset, options, value):
        """Apply `geo_polygon` query.

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return queryset.query(
            Q("geo_polygon", **cls.get_geo_polygon_params(value, options["field"]))
        )

    @classmethod
    def apply_query_geo_bounding_box(cls, queryset, options, value):
        """Apply `geo_bounding_box` query.

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return queryset.query(
            Q(
                "geo_bounding_box",
                **cls.get_geo_bounding_box_params(value, options["field"])
            )
        )

    @classmethod
    def apply_query_geo_shape(cls, queryset, options, value):
        """Apply `geo_shape` query.

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return queryset.query(
            Q("geo_shape", **cls.get_geo_shape_params(value, options["field"]))
        )

    def get_filter_query_params(self, request, view):
        """Get query params to be filtered on.

        :param request: Django REST framework request.
        :param view: View.
        :type request: rest_framework.request.Request
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Request query params to filter on.
        :rtype: dict
        """
        query_params = request.query_params.copy()

        filter_query_params = {}
        filter_fields = self.prepare_filter_fields(view)
        for query_param in query_params:
            query_param_list = self.split_lookup_filter(query_param, maxsplit=1)
            field_name = query_param_list[0]

            if field_name in filter_fields:
                lookup_param = None
                if len(query_param_list) > 1:
                    lookup_param = query_param_list[1]

                valid_lookups = filter_fields[field_name]["lookups"]

                if lookup_param is None or lookup_param in valid_lookups:
                    values = [
                        __value.strip()
                        for __value in query_params.getlist(query_param)
                        if __value.strip() != ""
                    ]

                    if values:
                        filter_query_params[query_param] = {
                            "lookup": lookup_param,
                            "values": values,
                            "field": filter_fields[field_name].get("field", field_name),
                            "type": view.mapping,
                        }
        return filter_query_params

    def filter_queryset(self, request, queryset, view):
        """Filter the queryset.

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        filter_query_params = self.get_filter_query_params(request, view)
        for options in filter_query_params.values():

            # For all other cases, when we don't have multiple values,
            # we follow the normal flow.
            for value in options["values"]:

                # `geo_distance` query lookup
                if options["lookup"] == LOOKUP_FILTER_GEO_DISTANCE:
                    queryset = self.apply_query_geo_distance(queryset, options, value)

                # `geo_polygon` query lookup
                elif options["lookup"] == LOOKUP_FILTER_GEO_POLYGON:
                    queryset = self.apply_query_geo_polygon(queryset, options, value)

                # `geo_bounding_box` query lookup
                elif options["lookup"] == LOOKUP_FILTER_GEO_BOUNDING_BOX:
                    queryset = self.apply_query_geo_bounding_box(
                        queryset, options, value
                    )

                # `geo_shape` query lookup
                elif options["lookup"] == LOOKUP_FILTER_GEO_SHAPE:
                    queryset = self.apply_query_geo_shape(queryset, options, value)

        return queryset



"""
Apps.
"""

from django.apps import AppConfig

__title__ = "django_elasticsearch_dsl_drf_alt.apps"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("Config",)


class Config(AppConfig):
    """Config."""

    name = "django_elasticsearch_dsl_drf_alt"
    label = "django_elasticsearch_dsl_drf_alt"



"""
Source backend.
"""

from rest_framework.filters import BaseFilterBackend

__title__ = "django_elasticsearch_dsl_drf_alt.source"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("SourceBackend",)


class SourceBackend(BaseFilterBackend):
    """Static source backend.

    Example 1 (simple):

        >>> from django_elasticsearch_dsl_drf_alt.filter_backends import (
        >>>     SourceBackend
        >>> )
        >>> from django_elasticsearch_dsl_drf_alt.viewsets import (
        >>>     BaseDocumentViewSet,
        >>> )
        >>>
        >>> # Local article document definition
        >>> from .documents import ArticleDocument
        >>>
        >>> # Local article document serializer
        >>> from .serializers import ArticleDocumentSerializer
        >>>
        >>> class ArticleDocumentView(BaseDocumentViewSet):
        >>>
        >>>     document = ArticleDocument
        >>>     serializer_class = ArticleDocumentSerializer
        >>>     filter_backends = [SourceBackend,]
        >>>     source = ["title"]

    Example 2 (complex):

        >>> # ...
        >>>     source = ["title", "author.*"]

    Example 3 (even more complex):

        >>> # ...
        >>>     source = {
        >>>         "includes": ["title", "author.*"],
        >>>         "excludes": [ "*.description" ]
        >>>     }

    Source can make queries lighter. However, it can break current
    functionality. Use it with caution.
    """

    def filter_queryset(self, request, queryset, view):
        """Filter the queryset.

        :param request: Django REST framework request.
        :param queryset: Base queryset.
        :param view: View.
        :type request: rest_framework.request.Request
        :type queryset: elasticsearch_dsl.search.Search
        :type view: rest_framework.viewsets.ReadOnlyModelViewSet
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        if getattr(view, "source", None) is not None:
            queryset = queryset.source(view.source)

        return queryset



import operator
import six

from elasticsearch_dsl.query import Q

from .base import BaseSearchQueryBackend

__title__ = (
    "django_elasticsearch_dsl_drf_alt.filter_backends.search." "query_backends.nested"
)
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2017-2020 Artur Barseghyan"
__license__ = "GPL 2.0/LGPL 2.1"
__all__ = ("NestedQueryBackend",)


class NestedQueryBackend(BaseSearchQueryBackend):
    """Nested query backend."""

    query_type = "nested"

    @classmethod
    def construct_search(cls, request, view, search_backend):
        """Construct search.

        Dictionary key is the GET param name. The path option stands for the
        path in Elasticsearch.

        Type 1:

            search_nested_fields = {
                'country': {
                    'path': 'country',
                    'fields': ['name'],
                },
                'city': {
                    'path': 'country.city',
                    'fields': ['name'],
                },
            }

        Type 2:

            search_nested_fields = {
                'country': {
                    'path': 'country',
                    'fields': [{'name': {'boost': 2}}]
                },
                'city': {
                    'path': 'country.city',
                    'fields': [{'name': {'boost': 2}}]
                },
            }

        :param request:
        :param view:
        :param search_backend:
        :return:
        """
        if not hasattr(view, "search_nested_fields"):
            return []

        query_params = search_backend.get_search_query_params(request)
        __queries = []

        for search_term in query_params:
            __values = search_backend.split_lookup_name(search_term, 1)
            __len_values = len(__values)
            if __len_values > 1:
                label, value = __values
                if label in view.search_nested_fields:
                    options = view.search_nested_fields.get(label)
                    path = options.get("path")

                    queries = []
                    for _field in options.get("fields", []):
                        # In case if we deal with structure 2
                        if isinstance(_field, dict):
                            # take options (ex: boost) into consideration
                            field_options = {
                                key: value
                                for key, value in _field.items()
                                if key != "name"
                            }
                            field_options.update(
                                {
                                    "query": search_term,
                                }
                            )
                            field = "{}.{}".format(path, _field["name"])
                            field_kwargs = {
                                field: field_options,
                            }
                        # In case if we deal with structure 1
                        else:
                            field = "{}.{}".format(path, _field)
                            field_kwargs = {field: value}

                        queries = [Q("match", **field_kwargs)]

                    __queries.append(
                        Q(
                            cls.query_type,
                            path=path,
                            query=six.moves.reduce(operator.or_, queries),
                        )
                    )
            else:
                for label, options in view.search_nested_fields.items():
                    queries = []
                    path = options.get("path")

                    for _field in options.get("fields", []):
                        # In case if we deal with structure 2
                        if isinstance(_field, dict):
                            # take options (ex: boost) into consideration
                            field_options = {
                                key: value
                                for key, value in _field.items()
                                if key != "name"
                            }
                            field_options.update(
                                {
                                    "query": search_term,
                                }
                            )
                            field = "{}.{}".format(path, _field["name"])
                            field_kwargs = {
                                field: field_options,
                            }
                        # In case if we deal with structure 1
                        else:
                            field = "{}.{}".format(path, _field)
                            field_kwargs = {field: search_term}

                        queries.append(Q("match", **field_kwargs))

                    __queries.append(
                        Q(
                            cls.query_type,
                            path=path,
                            query=six.moves.reduce(operator.or_, queries),
                        )
                    )

        return __queries
